<?php
/**
 * WP Smart Analytics provides insights into visitor analytics and provides security to your WordPress installation.
 *
 * @package wpsa
 *
 * Plugin Name: WP Smart Analytics
 * Plugin URI: http://www.vinodsebastian.com/plugins
 * Description: WP Smart Analytics provides insights into visitor analytics provides security to your WordPress installation from the comfort of your WordPress dashboard. This mobile freindly plugin uses only encrypted text files for logs and database for statistics. It has option to anonymize IP's and data is stored locally.
 * Version: 1.2.1
 * Author: Vinod Sebastian
 * Author URI: http://www.vinodsebastian.com
 * License URI: https://www.gnu.org/licenses/gpl-3.0.html
 * Created Date: 13/07/18
 * Updated Date: 23/02/22
 */

if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly.
}
require_once ABSPATH . 'wp-admin/includes/file.php';

if ( ! class_exists( 'WPSA' ) ) :
	/**
	 * Class for WP Smart Analytics
	 */
	class WPSA {
		const SCRIPT_DEBUG = false;
		const SLACK        = 16; // Ensures the colors generated are neither too light or dark.
		const SEPERATOR    = '|--|'; // Seperator used to demarcate entities in cookies and flat files.
		const COOKIE_NAME  = 'wpsa_visitor_cookie'; // Name of the visitor cookie.

		const ACCESS_URL       = 'http://api.ipstack.com/'; // IP Stack URL.
		const MAX_READ_URL     = 3; // How many failed attempts to read ACCESS_URL. Defaults to 3. Should be >= 1.
		const MAX_READ_TIMEOUT = 120; // Maximum read timeout ACCESS_URL.

		const ACCEPT_LANGUAGE_PRIORITY = .33; // Minimum priority for a preferred language of visitor.
		const ACCEPT_LANGUAGE_LIMIT    = 3; // Number of preferred languages of visitor.

		// Options.
		const SHA_SECRET_KEY              = 'SHA_SECRET_KEY'; // SHA Secret key.
		const SHA_SECRET_IV               = 'SHA_SECRET_IV'; // SHA Secret IV.
		const WPSA_API_KEY                = 'WPSA_API_KEY'; // IP Stack access key.
		const IGNORE_FILES                = 'IGNORE_FILES'; // What file path visits to ignore.
		const TAG_FREQUENCY               = 'TAG_FREQUENCY'; // How many user visits are required for a tag to be displayed?
		const SESSION_DURATION            = 'SESSION_DURATION'; // How much duration on a page can be counted as part of a session. In minutes.
		const COOKIE                      = 'COOKIE'; // Whether to ignore my browsing using cookie.
		const IP_STRING                   = 'IP_STRING'; // Whether to ignore my browsing using my IP.
		const IGNORE_BOTS                 = 'IGNORE_BOTS'; // Whether to ignore bots. No need to log them.
		const ANONYMIZE_IP                = 'ANONYMIZE_IP'; // Should we mask IP addresses.
		const STOP_LOGGING                = 'STOP_LOGGING'; // Should we stop logging in text file.
		const VISITOR_COUNT               = 'VISITOR_COUNT'; // Restart counting visitors. Start at 1.
		const REFRESH_TODAYS_LOG_DURATION = 'REFRESH_TODAYS_LOG_DURATION'; // Duration at which today's log is automatically refreshed.
		const MAX_CACHE_RECORDS           = 'MAX_CACHE_RECORDS'; // Maximum cache rows. Old ones are purged.
		const MAX_DATABASE_RECORDS        = 'MAX_DATABASE_RECORDS'; // Maximum database rows. Old ones are purged.

		/**
		 * Holds instance of class Browser_Detection.
		 *
		 * @var object $browser.
		 */
		private $browser;

		/**
		 * Refers to a single instance of this class.
		 *
		 * @var object $instance Instance.
		 */
		private static $instance;

		/**
		 * Registers the plugin using singleton pattern.
		 */
		public static function get_instance() {
			if ( null === self::$instance ) {
				self::$instance = new self();
			}

			return self::$instance;
		} // end get_instance;

		/**
		 * Initializes the plugin by setting localization, filters, and administration functions.
		 */
		private function __construct() {
			// Run this on plugin activation.
			register_activation_hook( __FILE__, array( $this, 'wp_smart_analytics_create_plugin_database_table' ) );

			$tag_frequency = get_option( self::TAG_FREQUENCY, 3 );
			update_option( self::TAG_FREQUENCY, $tag_frequency );

			$session_duration = get_option( self::SESSION_DURATION, 30 );
			update_option( self::SESSION_DURATION, $session_duration );

			$max_cache_records = get_option( self::MAX_CACHE_RECORDS, 10000 );
			update_option( self::MAX_CACHE_RECORDS, $max_cache_records );

			$max_database_records = get_option( self::MAX_DATABASE_RECORDS, 100000 );
			update_option( self::MAX_DATABASE_RECORDS, $max_database_records );

			global $wp_filesystem;
			WP_Filesystem();
			$upload_dir = wp_upload_dir();
			$wpsa_dir   = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics';
			if ( ! $wp_filesystem->is_dir( $wpsa_dir ) ) {
				$wp_filesystem->mkdir( $wpsa_dir );
			}
			if ( $wp_filesystem->is_dir( $wpsa_dir ) ) {
				chmod( $wpsa_dir, 0744 );
				$wp_filesystem->put_contents( $wpsa_dir . '/.htaccess', 'Deny from  all', 0644 );
			}
			$wpsa_log_dir = $wpsa_dir . '/log';
			if ( ! $wp_filesystem->is_dir( $wpsa_log_dir ) ) {
				$wp_filesystem->mkdir( $wpsa_log_dir );
			}
			if ( $wp_filesystem->is_dir( $wpsa_log_dir ) ) {
				chmod( $wpsa_log_dir, 0744 );
				$wp_filesystem->put_contents( $wpsa_log_dir . '/.htaccess', 'Deny from  all', 0644 );
			}
			$wpsa_meta_dir = $wpsa_dir . '/meta';
			if ( ! $wp_filesystem->is_dir( $wpsa_meta_dir ) ) {
				$wp_filesystem->mkdir( $wpsa_meta_dir );
			}
			if ( $wp_filesystem->is_dir( $wpsa_meta_dir ) ) {
				chmod( $wpsa_meta_dir, 0744 );
				$wp_filesystem->put_contents( $wpsa_meta_dir . '/.htaccess', 'Deny from  all', 0644 );
			}

			$secret_key = get_option( self::SHA_SECRET_KEY, substr( str_shuffle( '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ' ), 1, 32 ) ); // User defined private key.
			update_option( self::SHA_SECRET_KEY, $secret_key );

			$secret_iv = get_option( self::SHA_SECRET_IV, substr( str_shuffle( '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ' ), 1, 32 ) ); // User defined secret key.
			update_option( self::SHA_SECRET_IV, $secret_iv );

			add_action( 'admin_menu', array( $this, 'create_admin_menu' ) );

			add_action( 'wp_ajax_wpsa_view_log', array( $this, 'wp_ajax_wpsa_view_log' ) );
			add_action( 'wp_ajax_wpsa_delete_log', array( $this, 'wp_ajax_wpsa_delete_log' ) );
			add_action( 'wp_ajax_wpsa_add_tag', array( $this, 'wp_ajax_wpsa_add_tag' ) );
			add_action( 'wp_ajax_wpsa_save_settings', array( $this, 'wp_ajax_wpsa_save_settings' ) );
			add_action( 'wp_ajax_wpsa_generate_report', array( $this, 'wp_ajax_wpsa_generate_report' ) );

			if ( true === $this->is_frontend() ) {
					add_action( 'init', array( $this, 'ban_ip' ) );
					add_action( 'init', array( $this, 'set_cookie_session' ) );
					add_action( 'wp_footer', array( $this, 'log_visitor' ) );
			} elseif ( is_admin() ) {
				add_action( 'init', array( $this, 'detect_admin_cookie' ) );
			}

			$this->register_shortcode();
		} // end constructor

		/**
		 * --------------------------------Craete Database Section-----------------------------------------.
		 */
		/**
		 * Adds menu item called WP Smart Analytics.
		 *
		 * @return void
		 */
		public function wp_smart_analytics_create_plugin_database_table() {
			global $wpdb;
			$charset_collate = $wpdb->get_charset_collate();
			$table_name      = $wpdb->prefix . 'smart_analytics_visitor_log';

			$sql = "CREATE TABLE IF NOT EXISTS $table_name (
				id int NOT NULL AUTO_INCREMENT,
				time datetime DEFAULT CURRENT_TIMESTAMP NOT NULL,
				ip varchar(39) NOT NULL,
				ip_version varchar(15) NOT NULL,
				url varchar(255) NOT NULL,
				page_type varchar(63) NOT NULL,
				title varchar(255) NOT NULL,
				name varchar(127) NOT NULL,
				address varchar(255) NOT NULL,
				country varchar(128) NOT NULL,
				bot_info varchar(63) NOT NULL,
				platform varchar(63) NOT NULL,
				browser_name varchar(63) NOT NULL,
				browser_version varchar(63) NOT NULL,
				accept_languages varchar(255) NOT NULL,
				referrer_url varchar(255) NOT NULL,
				referrer_source varchar(127) NOT NULL,
				is_bot boolean NOT NULL,
				is_admin boolean NOT NULL,
				time_in_sec smallint(5) NOT NULL,
				UNIQUE KEY id (id)
			) $charset_collate;";

			require_once ABSPATH . 'wp-admin/includes/upgrade.php';
			dbDelta( $sql );
		}

		/**
		 * --------------------------------Admin Section-----------------------------------------.
		 */
		/**
		 * Adds menu item called WP Smart Analytics.
		 *
		 * @return void
		 */
		public function create_admin_menu() {
			add_action( 'admin_enqueue_scripts', array( $this, 'register_admin_assets' ) );

			if ( current_user_can( 'manage_options' ) ) {
				// Manage access.
				add_menu_page( 'WP Smart Analytics', 'WP Smart Analytics', 'manage_options', 'wp_smart_analytics', array( $this, 'manage_plugin' ), 'dashicons-info', 98 );
			} elseif ( current_user_can( 'read' ) ) {
				// Read access.
				add_menu_page( 'WP Smart Analytics', 'WP Smart Analytics', 'read', 'wp_smart_analytics', array( $this, 'read_plugin' ), 'dashicons-info', 98 );
			}
		}

		/**
		 * Registers admin assets.
		 *
		 * @return void
		 */
		public function register_admin_assets() {
			$screen = get_current_screen();

			if ( null === $screen ) {
				return;
			}

			if ( in_array( $screen->id, array( 'toplevel_page_wp_smart_analytics' ), true ) ) {
				$min = defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG ? '' : '.min';

				$my_js_url                = plugin_dir_url( __FILE__ ) . "assets/js/wpsa{$min}.js";
				$my_css_toggle_switch_url = plugin_dir_url( __FILE__ ) . 'assets/css/toggle-switch.css';
				$my_css_url               = plugin_dir_url( __FILE__ ) . 'assets/css/wpsa.css';

				$my_js_path                = plugin_dir_path( __FILE__ ) . "/assets/js/wpsa{$min}.js";
				$my_css_toggle_switch_path = plugin_dir_path( __FILE__ ) . '/assets/css/toggle-switch.css';
				$my_css_path               = plugin_dir_path( __FILE__ ) . '/assets/css/wpsa.css';

				$my_js_version                = gmdate( 'ymd-Gis', filemtime( $my_js_path ) );
				$my_css_toggle_switch_version = gmdate( 'ymd-Gis', filemtime( $my_css_toggle_switch_path ) );
				$my_css_version               = gmdate( 'ymd-Gis', filemtime( $my_css_path ) );

				// Scripts.
				$wpsc_js = wp_register_script( 'wpsa_js', $my_js_url, array( 'jquery', 'jquery-ui-core' ), $my_js_version, true );
				wp_localize_script(
					'wpsa_js',
					'ajax_var',
					array(
						'url'   => admin_url( 'admin-ajax.php' ),
						'nonce' => wp_create_nonce( 'wpsa_nonce' ),
					)
				);
				wp_enqueue_script( 'wpsa_js', $my_js_url, array( 'jquery', 'jquery-ui-core' ), $my_js_version, true );
				wp_enqueue_script( 'jquery-ui-datepicker' );

				// Styles.
				wp_register_style( 'wpsa_css_toggle_switch', $my_css_toggle_switch_url, array(), $my_css_toggle_switch_version );
				wp_enqueue_style( 'wpsa_css_toggle_switch', $my_css_toggle_switch_url, array(), $my_css_toggle_switch_version );

				wp_register_style( 'wpsa_css', $my_css_url, array( 'wpsa_css_toggle_switch' ), $my_css_version );
				wp_enqueue_style( 'wpsa_css', $my_css_url, array( 'wpsa_css_toggle_switch' ), $my_css_version );

				$jquery_ui_css_version = gmdate( 'ymd-Gis', filemtime( plugin_dir_path( __FILE__ ) . 'assets/css/jquery-ui.structure.css' ) );
				wp_register_style( 'jquery-ui', plugin_dir_url( __FILE__ ) . 'assets/css/jquery-ui.structure.css', array(), $jquery_ui_css_version );
				wp_enqueue_style( 'jquery-ui', plugin_dir_url( __FILE__ ) . 'assets/css/jquery-ui.structure.css', array(), $jquery_ui_css_version );

				$jquery_ui_css_custom_calendar_version = gmdate( 'ymd-Gis', filemtime( plugin_dir_path( __FILE__ ) . 'assets/css/jquery-ui.theme.css' ) );
				wp_register_style( 'jquery-ui-custom-calendar', plugin_dir_url( __FILE__ ) . 'assets/css/jquery-ui.theme.css', array( 'jquery-ui' ), $jquery_ui_css_custom_calendar_version );
				wp_enqueue_style( 'jquery-ui-custom-calendar', plugin_dir_url( __FILE__ ) . 'assets/css/jquery-ui.theme.css', array( 'jquery-ui' ), $jquery_ui_css_custom_calendar_version );
			}
		}

		/**
		 * Manage Plugin.
		 *
		 * @return void
		 */
		public function manage_plugin() {
			$plugin_data    = get_plugin_data( __FILE__ );
			$plugin_version = $plugin_data['Version'];

			$str = "<div id='wpsa'>";
			$str = $str . "<script type='text/javascript'>let timer = null;</script>";
			$str = $str . "<div id='toTop'>To The Top!</div>";
			$str = $str . "<div id='meta_tag' class='modal'><span>Please enter the tag below.</span><input type='text' id='tag'/><input type='hidden' id='ip'/><p style='float:left;'><a class='button-primary' rel='noopener noreferrer' onclick='save_tag()'>&nbsp;Save&nbsp;</a></p><p style='float:left; margin-left:20px;'><a class='button-primary' rel='noopener noreferrer' onclick='jQuery(this).parent().parent().hide(); cancel_tag();'>Cancel</a></p></div>";
			$str = $str . "<div class='banner table' style='margin-top:24px;'><img src='" . plugin_dir_url( __FILE__ ) . "assets/images/icon.png' style='clear:both; float:left; width:64px;' /><h1 style='clear:both; float:left; line-height:1.16em; width:100%; margin:0px;' class='blue'><span style='float:left;'>" . __( 'WP Smart Analytics', 'wpsa' ) . "&nbsp;</span><span style='float:left; font-size:14px; font-size:.75em; font-weight:600; line-height:1.85em;'>" . 'v' . $plugin_version . "</span> <span style='clear:both; float:left; font-size:14px;'>by Vinod Sebastian&nbsp;</span><span style='float:left; font-size:14px;'>(Support: business@vinodsebastian.com)</span></h1><p style='float:right; margin-left:10px; margin-right:10px;'><a target='_blank' class='button-primary' rel='noopener noreferrer' href='https://wordpress.org/plugins/wp-smart-analytics/#reviews'>OK, I'll leave a review</a></p><p style='float:right; margin-left:10px; margin-right:10px;'><a target='_blank' class='button-primary' rel='noopener noreferrer' href='https://wordpress.org/support/plugin/wp-smart-analytics/'>Need some support?</a></p></div>";
			$str = $str . "<div class='wrap'>";
			$str = $str . "<span id='wpsamain' style='clear:both; float:left; width:100%;'></span>";
			$str = $str . "<h2 style='clear:both; float:left; width:100%;'></h2>";
			$str = $str . "<div style='clear:both; float:left; width:100%;' id='plugin-messages'></div>";
			$str = $str . "<div class='main section'>";
			$str = $str . "<div class='header'><h2><span style='float:left; width:32px;' class='dashicons dashicons-admin-settings'></span><span  style='float:left;' >" . __( 'Settings.', 'wpsa' ) . "</span><a class='max' style='float:right; text-decoration:none;' rel='noopener noreferrer' onclick='open_section(jQuery(this));'><span class='dashicons dashicons-plus'></span></a><a class='min' style='float:right; text-decoration:none;' rel='noopener noreferrer' onclick='close_section(jQuery(this));'><span class='dashicons dashicons-minus'></span></a></h2></div>";
			$str = $str . "<div class='collapse'>";
			$str = $str . '<h3>' . __( "Provide ipstack.com <a target='_blank' class='blue' href='https://ipstack.com/signup/free'>API key</a> (It's got free API service used to obtain locality of an IP address).<br>The plugin gracefully degrades without it.", 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->api_key();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Do not log (File paths containing any one of the comma seperated strings below are ignored).', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->ignore_files();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Minimum visitor frequency for a tag to be created.', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->tag_frequency();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Maximum duration visitor can stay on a page as part of an ongoing session (In minutes).', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->session_duration();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Meta tags.', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->meta_tags();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Ignore my visits (Using cookies).', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->ignore_my_visits_cookie();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Ignore my visits (Using IP).', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->ignore_my_visits_ip();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Ignore bot visits (Visits from Googlebot, Bingbot and so on are not logged).', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->ignore_bots();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Anonymize IP (This will mask last 3 digits of IP address). This is recommended but would reduce accuracy and tagging and banning cannot be done.', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->anonymize_ip();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Do not log visitor paths in flat files.', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->stop_logging();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Reset my visitor count. This is not recommended.', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->reset_visitor_count();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( "Duration of log refresh in minutes. It can be decimal. This affects only when viewing today's logs. A change in value is affected only after page reload. This helps in viewing logs real time. This setting is common for admin and [wpsa_log] short code screen.", 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->refresh_todays_log_duration();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Maximum limit to visitor records in cache.', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->max_cache_records();
			$str = $str . '</div>';
			$str = $str . '<h3>' . __( 'Maximum limit to visitor records in database. The database is immediately truncated on save.', 'wpsa' ) . '</h3>';
			$str = $str . '<div>';
			$str = $str . $this->max_database_records();
			$str = $str . '</div>';
			$str = $str . "<a id='wpsa_ajax_save_setting_button' class='button-primary' style='margin-top:20px;' rel='noopener noreferrer' onclick=\"save_settings();\">&nbsp;Save&nbsp;</a><a id='wpsa_ajax_cancel_setting_button' class='button-primary' style='margin-top:20px; margin-left:20px;' rel='noopener noreferrer' onclick=\"close_section(jQuery(this).parent()); jQuery('html, body').animate({scrollTop: jQuery('#wpsamain').offset().top - 30}, 100, 'swing');\">Cancel</a>";
			$str = $str . '</div>';
			$str = $str . '</div>';

			$str = $str . "<div class='main section'>";
			$str = $str . "<div class='header'><h2><span style='float:left; width:32px;' class='dashicons dashicons-analytics'></span><span  style='float:left;' >" . __( 'Report.', 'wpsa' ) . "</span><a class='max' style='float:right; text-decoration:none;' rel='noopener noreferrer' onclick='open_section(jQuery(this));'><span class='dashicons dashicons-plus'></span></a><a class='min' style='float:right; text-decoration:none;' rel='noopener noreferrer' onclick='close_section(jQuery(this));'><span class='dashicons dashicons-minus'></span></a></h2></div>";
			$str = $str . "<div class='collapse'>";
			$str = $str . '<div style="clear:both; float:left; width:100%;">';
			$str = $str . '<div style="clear:both; float:left; width:25%; text-align:right; padding-left:1%; padding-right:1%; font-size:1.33em; line-height:2.33em; min-width:100px;"><sup style="color:red;">*</sup>Period (Incl.):</div>';
			$str = $str . '<div style="float:left; width:25%; text-align:left; padding-left:2%; padding-right:2%; min-width:250px;"><input style="float:left; width:100%; margin-top:2px;" type="text" class="wp_smart_analytics_date" name="start_date" id="start_date" value=""/></div>';
			$str = $str . '<div style="float:left; width:25%; text-align:left; padding-left:2%; padding-right:2%; min-width:250px;"><input style="float:left; width:100%; margin-top:2px;"  type="text" class="wp_smart_analytics_date" name="end_date" id="end_date" value=""/></div>';
			$str = $str . '</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">&nbsp;</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">';
			$str = $str . '<div style="clear:both; float:left; width:25%; text-align:right; padding-left:1%; padding-right:1%; font-size:1.33em; line-height:1.66em; min-width:100px;"><sup style="color:red;">*</sup>Category:</div>';
			$str = $str . '<select id="category_group" style="float:left; width:25%; text-align:left; margin-left:2%; padding-top:4px; padding-bottom:4px; min-width:250px; border-width:2px; padding-left:4px;">';
			$str = $str . '<option value="page_type">Page Type</option>';
			$str = $str . '<option value="title" selected="selected">Page Title</option>';
			$str = $str . '<option value="name">Visitor Name</option>';
			$str = $str . '<option value="country">Visitor Country</option>';
			$str = $str . '<option value="platform">Visitor Platform</option>';
			$str = $str . '<option value="browser_name">Visitor Browser</option>';
			$str = $str . '<option value="referrer_source">Referrer Source</option>';
			$str = $str . '</select>';
			$str = $str . '</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">&nbsp;</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">';
			$str = $str . '<div style="clear:both; float:left; width:25%; text-align:right; padding-left:1%; padding-right:1%; font-size:1.33em; line-height:1.33em; min-width:100px;"><sup style="color:red;">*</sup>Remove Bot Visits:</div>';
			$str = $str . '<div style="float:left; width:25%; text-align:left; padding-left:2%; padding-right:2%; min-width:250px; padding-top:8px;">';
			$str = $str . "<label for='remove_bot_visits' class='switch'>";
			$str = $str . "<input type='checkbox' id='remove_bot_visits' checked='checked' />";
			$str = $str . "<span class='remove_bot_visits slider round'></span>";
			$str = $str . '</label>';
			$str = $str . '</div>';
			$str = $str . '</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">&nbsp;</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">';
			$str = $str . '<div style="clear:both; float:left; width:25%; text-align:right; padding-left:1%; padding-right:1%; font-size:1.33em; line-height:1.33em; min-width:100px;"><sup style="color:red;">*</sup>Remove Admin Visits:</div>';
			$str = $str . '<div style="float:left; width:25%; text-align:left; padding-left:2%; padding-right:2%; min-width:250px; padding-top:8px;">';
			$str = $str . "<label for='remove_admin_visits' class='switch'>";
			$str = $str . "<input type='checkbox' id='remove_admin_visits' checked='checked' />";
			$str = $str . "<span class='remove_admin_visits slider round'></span>";
			$str = $str . '</label>';
			$str = $str . '</div>';
			$str = $str . '</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">&nbsp;</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">';
			$str = $str . '<div style="clear:both; float:left; width:25%; text-align:right; padding-left:1%; padding-right:1%; font-size:1.33em; line-height:1.66em; min-width:100px;"><sup style="color:red;">*</sup>Sort By:</div>';
			$str = $str . '<select id="sort" style="float:left; width:25%; text-align:left; margin-left:2%; padding-top:4px; padding-bottom:4px; min-width:250px; border-width:2px; padding-left:4px;">';
			$str = $str . '<option value="category_group">Category</option>';
			$str = $str . '<option value="visitor_count" selected="selected">Visitor Count</option>';
			$str = $str . '<option value="sum_page_load_count">Page Loads</option>';
			$str = $str . '<option value="sum_duration_in_sec">Duration (In Sec)</option>';
			$str = $str . '</select>';
			$str = $str . '</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">&nbsp;</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">';
			$str = $str . '<div style="clear:both; float:left; width:25%; text-align:right; padding-left:1%; padding-right:1%; font-size:1.33em; line-height:1.66em; min-width:100px;"><sup style="color:red;">*</sup>Direction:</div>';
			$str = $str . '<select id="direction" style="float:left; width:25%; text-align:left; margin-left:2%; padding-top:4px; padding-bottom:4px; min-width:250px; border-width:2px; padding-left:4px;">';
			$str = $str . '<option value="ASC">Ascending</option>';
			$str = $str . '<option value="DESC" selected="selected">Descending</option>';
			$str = $str . '</select>';
			$str = $str . '</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">&nbsp;</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">';
			$str = $str . '<div style="clear:both; float:left; width:25%; text-align:right; padding-left:1%; padding-right:1%; font-size:1.33em; line-height:1.33em; min-width:100px;"><sup style="color:red;">&nbsp;</div>';
			$str = $str . '<div style="float:left; width:25%; text-align:left; padding-left:2%; padding-right:2%; min-width:250px; padding-top:8px;">';
			$str = $str . "<a id='wpsa_ajax_generate-report_button' class='button-primary' style='margin-top:20px;' rel='noopener noreferrer' onclick=\"generate_report();\">&nbsp;Generate&nbsp;</a><a id='wpsa_ajax_cancel_report_button' class='button-primary' style='margin-top:20px; margin-left:20px;' rel='noopener noreferrer' onclick=\"close_section(jQuery(this).parent()); jQuery('html, body').animate({scrollTop: jQuery('#wpsamain').offset().top - 30}, 100, 'swing');\">Cancel</a>";
			$str = $str . '</div>';
			$str = $str . '</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">&nbsp;</div>';
			$str = $str . '<div style="clear:both; float:left; width:100%;">&nbsp;</div>';
			$str = $str . "<table class='features-table'>";
			$str = $str . '<thead>';
			$str = $str . '<tr>';
			$str = $str . '<td style="font-weight:bold; width:5%; overflow-x:hidden;">S. NO.</td>';
			$str = $str . '<td style="font-weight:bold; width:65%;">CATEGORY</td>';
			$str = $str . '<td style="font-weight:bold; width:10%; min-width:20px; overflow-x:hidden;">VISITOR COUNT</td>';
			$str = $str . '<td style="font-weight:bold; width:10%; min-width:20px; overflow-x:hidden;">PAGE LOADS</td>';
			$str = $str . '<td style="font-weight:bold; width:10%; min-width:20px; overflow-x:hidden;">DURATION (IN SEC)</td>';
			$str = $str . '</tr>';
			$str = $str . '</thead>';
			$str = $str . '<tbody id="table_message">';
			$str = $str . '<tr><td colspan = "6">No data has been generated yet ...</td></tr>';
			$str = $str . '</tbody>';
			$str = $str . '</table>';
			$str = $str . '</div>';
			$str = $str . '</div>';
			$str = $str . '</div>';

			$str          = $str . "<div id='sidebar' class='section log_section'>";
			$str          = $str . "<div class='header'><h2><span style='float:left; width:32px;' class='dashicons dashicons-text-page'></span><span  style='float:left;' >" . __( 'Logs.', 'wpsa' ) . "</span><a class='max' style='float:right; text-decoration:none;' rel='noopener noreferrer' onclick='open_logs();'><span class='dashicons dashicons-plus'></span></a><a class='min' style='float:right; text-decoration:none;' rel='noopener noreferrer' onclick='close_logs();'><span class='dashicons dashicons-minus'></span></a></h2></div>";
			$str          = $str . "<div class='collapse1'>";
			$str          = $str . '<p>' . __( 'Click on them to either view or delete.', 'wpsa' ) . '</p>';
			$str          = $str . '<table>';
			$upload_dir   = wp_upload_dir();
			$wpsa_log_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/log/';
			$files        = $this->get_files( $wpsa_log_dir );
			$cnt          = 0;
			foreach ( $files as $file ) {
				$cnt++;
				$str = $str . "<tr class='li' id='li_" . str_replace( '_', '', str_replace( '.', '', $file ) ) . "'>";
				$str = $str . "<td class='id'>{$cnt}</td><td class='filename' " . ( ( 1 === $cnt ) ? "style='font-weight:bold;'" : '' ) . '>' . str_replace( '_', ' ', str_replace( '.txt', '', $file ) ) . "</td><td class='action'><a onclick=\"ajax_view_log('" . $file . "');\">VIEW</a></td><td class='action'><a onclick=\"ajax_delete_log('" . $file . "');\">DELETE</a></td>";
				$str = $str . '</tr>';
			}
			$str = $str . '</table>';
			$str = $str . '</div>';
			$str = $str . '</div>';

			$str = $str . "<div class='section_right log_section collapse2'>";
			$str = $str . '<h2>' . __( 'Log Details.', 'wpsa' ) . '</h2>';
			$str = $str . "<p id='loading'>" . __( 'Loading', 'wpsa' ) . ' ...</p>';
			$str = $str . '<p>' . __( 'The cloud tag below contains all tags with frequency', 'wpsa' ) . ' ' . get_option( self::TAG_FREQUENCY ) . ' ' . __( 'or more. You can click on them to view its refined logs.', 'wpsa' ) . '</p>';
			$str = $str . '<p>' . __( "If you want to tag a visitor, click on the person with the ball and provide the tag name. If you give the name 'ban' they would get a 403 Access Forbidden message when accessing WordPress front-end.", 'wpsa' ) . '</p>';

			$str = $str . '</div>';

			$str = $str . "<div id='log' class='log_section collapse3'>";
			$str = $str . $this->view_logs( 'manage_options' );
			$str = $str . '</div>';
			$str = $str . "<script type='text/javascript'>window.location.hash ='#wpsamain';</script>";
			$str = $str . '</div>';

			echo $str;
		}

		/**
		 * Read Plugin.
		 *
		 * @return void
		 */
		public function read_plugin() {
			$plugin_data    = get_plugin_data( __FILE__ );
			$plugin_version = $plugin_data['Version'];

			$str          = "<div id='wpsa'>";
			$str          = $str . "<script type='text/javascript'>let timer = null;</script>";
			$str          = $str . "<div id='toTop'>To The Top!</div>";
			$str          = $str . "<div class='banner table' style='margin-top:24px; padding-bottom:5px;'><img src='" . plugin_dir_url( __FILE__ ) . "assets/images/icon.png' style='clear:both; float:left; width:64px;' /><h1 style='clear:both; float:left; line-height:1.16em; width:100%; margin:0px;' class='blue'><span style='float:left;'>" . __( 'WP Smart Analytics', 'wpsa' ) . "&nbsp;</span><span style='float:left; font-size:14px; font-size:.75em; font-weight:600; line-height:1.85em;'>" . 'v' . $plugin_version . "</span> <span style='clear:both; float:left; font-size:14px;'>by Vinod Sebastian&nbsp;</span><span style='float:left; font-size:14px;'>(Support: business@vinodsebastian.com)</span></h1><p style='float:right; margin-left:10px; margin-right:10px;'><a target='_blank' class='button-primary' rel='noopener noreferrer' href='https://wordpress.org/plugins/wp-smart-analytics/#reviews'>OK, I'll leave a review</a></p><p style='float:right; margin-left:10px; margin-right:10px;'><a target='_blank' class='button-primary' rel='noopener noreferrer' href='https://wordpress.org/support/plugin/wp-smart-analytics/'>Need some support?</a></p></div>";
			$str          = $str . "<div class='wrap'>";
			$str          = $str . "<h2 style='clear:both; float:left; width:100%; margin-top:25px; padding-bottom:5px;'></h2>";
			$str          = $str . "<div class='section_left'>";
			$str          = $str . '<h2>' . __( 'Logs.', 'wpsa' ) . '</h2>';
			$str          = $str . '<p>' . __( 'Click to view relevant log details.', 'wpsa' ) . '</p>';
			$str          = $str . '<table>';
			$upload_dir   = wp_upload_dir();
			$wpsa_log_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/log/';
			$files        = $this->get_files( $wpsa_log_dir );
			$cnt          = 0;
			foreach ( $files as $file ) {
				$cnt++;
				$str = $str . "<tr class='li' id='li_" . str_replace( '_', '', str_replace( '.', '', $file ) ) . "'>";
				$str = $str . "<td class='id'>{$cnt}</td><td class='filename' " . ( ( 1 === $cnt ) ? "style='font-weight:bold;'" : '' ) . '>' . str_replace( '_', ' ', str_replace( '.txt', '', $file ) ) . "</td><td class='action'><a onclick=\"ajax_view_log('" . $file . "');\">VIEW</a></td>";
				$str = $str . '</tr>';
			}
			$str = $str . '</table>';
			$str = $str . '</div>';

			$str = $str . "<div class='section_right'>";
			$str = $str . '<h2>' . __( 'Log Details.', 'wpsa' ) . '</h2>';
			$str = $str . "<p id='loading'>" . __( 'Loading.', 'wpsa' ) . '</p>';
			$str = $str . '<p>' . __( 'The cloud tag below contains all tags with frequency', 'wpsa' ) . ' ' . get_option( self::TAG_FREQUENCY ) . ' ' . __( 'or more. You can click on them to view its refined logs.', 'wpsa' ) . '</p>';
			$str = $str . '</div>';

			$str = $str . "<div id='log'>";
			$str = $str . $this->view_logs( 'read' );
			$str = $str . '</div>';
			$str = $str . '</div>';
			$str = $str . '</div>';

			echo $str;
		}


		/**
		 * Bans 'IP's.
		 *
		 * @return void
		 */
		public function ban_ip() {
			$incoming_ip = $this->get_ip_address();

			$tagarr = array();
			global $wp_filesystem;
			WP_Filesystem();
			$upload_dir    = wp_upload_dir();
			$wpsa_meta_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/meta';
			$path          = $wpsa_meta_dir . '/tags.txt';
			if ( file_exists( $path ) ) {
				$taglinestr = $wp_filesystem->get_contents( $path );
				$taglinestr = $this->encrypt_decrypt( $taglinestr, 'decrypt' );

				$taglinearr = explode( PHP_EOL, $taglinestr );
				foreach ( $taglinearr as $tagline ) {
					$tagline = trim( $tagline );
					if ( false !== stripos( $tagline, '=' ) ) {
						$tagarr   = explode( '=', $tagline );
						$tag_ip   = $tagarr[0];
						$tag_name = $tagarr[1];
						foreach ( $tagarr as $tagitem ) {
							if ( ( $incoming_ip === $tag_ip ) && ( 'ban' === trim( $tagitem ) ) ) {
								header( 'HTTP/1.0 403 Forbidden' );
								$str = '<!DOCTYPE html>';
								$str = $str . '<html lang="en">';
								$str = $str . '<head>';
								$str = $str . '  <meta charset="UTF-8">';
								$str = $str . '  <meta name="viewport" content="width=device-width, initial-scale=1.0">';
								$str = $str . '  <meta http-equiv="X-UA-Compatible" content="ie=edge">';
								$str = $str . '  <title>' . __( '403', 'wpsa' ) . '</title>';
								$str = $str . '  <style>';
								$str = $str . '		body{';
								$str = $str . '		  background: #233142;';
								$str = $str . '		}';
								$str = $str . '		h1{';
								$str = $str . '		  margin-top: 200px;';
								$str = $str . '		  margin-bottom: 20px;';
								$str = $str . '		  color: #00a0c8;';
								$str = $str . '		  text-align: center;';
								$str = $str . '		  font-size: 120px;';
								$str = $str . '		  font-weight: 1000;';
								$str = $str . '		}';
								$str = $str . '		h2{';
								$str = $str . '		  color: #455d7a;';
								$str = $str . '		  text-align: center;';
								$str = $str . '		  font-size: 30px;';
								$str = $str . '		  text-transform: uppercase;';
								$str = $str . '		}';
								$str = $str . '  </style>';
								$str = $str . '</head>';
								$str = $str . '<body>';
								$str = $str . '<h1>' . __( '403', 'wpsa' ) . '</h1>';
								$str = $str . '<h2>' . __( 'Not this time, access forbidden!', 'wpsa' ) . '</h2>';
								$str = $str . '</body>';
								$str = $str . '</html>';
								echo $str; // Can safely ignore.
								exit;
							}
						}
					}
				}
			}
		}

		/**
		 * Sets cookie
		 *
		 * @return void
		 */
		public function set_cookie_session() {
			if ( ! session_id() ) {
				session_start();
			}

			$ip_address = $this->get_ip_address();

			$time_from_file            = 0;
			$visitor_name_from_file    = '';
			$visitor_address_from_file = '';
			if ( '' !== $ip_address ) {
				$cache_from_file = $this->get_from_cache( $ip_address );

				if ( false === $cache_from_file ) {
					$ip_location               = $this->get_ip_location_details( $ip_address );
					$visitor_address_from_file = $this->get_visitor_location_from_ip_location_details( $ip_location );
				} else {
					$time_from_file            = $cache_from_file[1];
					$visitor_name_from_file    = $cache_from_file[2];
					$visitor_address_from_file = $cache_from_file[3];
				}
			}

			$cookie_name = self::COOKIE_NAME;

			$cookie_invalid = true; // Initially assuming cookie is invalid.
			if ( isset( $_COOKIE[ $cookie_name ] ) && ! empty( $_COOKIE[ $cookie_name ] ) ) { // Equivalent to else case below.
				$cookie_value = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_COOKIE[ $cookie_name ] ), FILTER_UNSAFE_RAW ) );
				$cookie_value = $this->encrypt_decrypt( $cookie_value, 'decrypt' );
				if ( 3 === substr_count( $cookie_value, self::SEPERATOR ) ) {
					$cookie_invalid = false;
				}
			}

			if ( ! isset( $_COOKIE[ $cookie_name ] ) || empty( $_COOKIE[ $cookie_name ] ) || $cookie_invalid ) {
				$session_ip              = $ip_address;
				$session_time            = $time_from_file;
				$session_visitor_name    = $this->generate_visitorname( $visitor_name_from_file );
				$session_visitor_address = $visitor_address_from_file;

				$session_value            = $session_ip . self::SEPERATOR . $session_time . self::SEPERATOR . $session_visitor_name . self::SEPERATOR . $session_visitor_address;
				$session_value            = $this->encrypt_decrypt( $session_value, 'encrypt' );
				$_SESSION[ $cookie_name ] = $session_value;

				$cookie_time = time();

				$cookie_value = $session_ip . self::SEPERATOR . $cookie_time . self::SEPERATOR . $session_visitor_name . self::SEPERATOR . $session_visitor_address;
				$cookie_value = $this->encrypt_decrypt( $cookie_value, 'encrypt' );
				setcookie( $cookie_name, $cookie_value, time() + 365 * DAY_IN_SECONDS, COOKIEPATH, COOKIE_DOMAIN );

				$this->put_to_cache( $ip_address, $session_visitor_name, $session_visitor_address );
			} else {
				$cookie_value = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_COOKIE[ $cookie_name ] ), FILTER_UNSAFE_RAW ) );
				$cookie_value = $this->encrypt_decrypt( $cookie_value, 'decrypt' );

				$cookie_array = explode( self::SEPERATOR, $cookie_value );

				$session_ip              = $ip_address;
				$session_time            = $cookie_array[1];
				$session_visitor_name    = $this->generate_visitorname( $cookie_array[2] );
				$session_visitor_address = $visitor_address_from_file;

				$session_value            = $session_ip . self::SEPERATOR . $session_time . self::SEPERATOR . $session_visitor_name . self::SEPERATOR . $session_visitor_address;
				$session_value            = $this->encrypt_decrypt( $session_value, 'encrypt' );
				$_SESSION[ $cookie_name ] = $session_value;

				$cookie_time = time();

				$cookie_value = $session_ip . self::SEPERATOR . $cookie_time . self::SEPERATOR . $session_visitor_name . self::SEPERATOR . $session_visitor_address;
				$cookie_value = $this->encrypt_decrypt( $cookie_value, 'encrypt' );
				setcookie( $cookie_name, $cookie_value, time() + 365 * DAY_IN_SECONDS, COOKIEPATH, COOKIE_DOMAIN );

				$this->put_to_cache( $ip_address, $session_visitor_name, $session_visitor_address );
			}
		}

		/**
		 * Sets cookie when admin screen opened.
		 * This enables cookie to alter visitor name from 'Visitor' to 'Admin'
		 *
		 * @return void
		 */
		public function detect_admin_cookie() {
			if ( current_user_can( 'read' ) ) {
				$ip_address = $this->get_ip_address();

				$time_from_file            = 0;
				$visitor_name_from_file    = '';
				$visitor_address_from_file = '';
				if ( '' !== $ip_address ) {
					$cache_from_file = $this->get_from_cache( $ip_address );

					if ( false === $cache_from_file ) {
						$ip_location               = $this->get_ip_location_details( $ip_address );
						$visitor_address_from_file = $this->get_visitor_location_from_ip_location_details( $ip_location );
					} else {
						$visitor_name_from_file    = $cache_from_file[2];
						$visitor_address_from_file = $cache_from_file[3];
					}
				}

				$cookie_name = self::COOKIE_NAME;

				$cookie_invalid = true; // Initially assuming cookie is invalid.
				if ( isset( $_COOKIE[ $cookie_name ] ) && ! empty( $_COOKIE[ $cookie_name ] ) ) { // Equivalent to else case below.
					$cookie_value = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_COOKIE[ $cookie_name ] ), FILTER_UNSAFE_RAW ) );
					$cookie_value = $this->encrypt_decrypt( $cookie_value, 'decrypt' );
					if ( 3 === substr_count( $cookie_value, self::SEPERATOR ) ) {
						$cookie_invalid = false;
					}
				}

				if ( ! isset( $_COOKIE[ $cookie_name ] ) || empty( $_COOKIE[ $cookie_name ] ) || $cookie_invalid ) {
					$cookie_ip              = $ip_address;
					$cookie_time            = 0;
					$cookie_visitor_name    = $this->generate_visitorname( $visitor_name_from_file );
					$cookie_visitor_address = $visitor_address_from_file;

					$cookie_value = $cookie_ip . self::SEPERATOR . $cookie_time . self::SEPERATOR . $cookie_visitor_name . self::SEPERATOR . $cookie_visitor_address;
					$cookie_value = $this->encrypt_decrypt( $cookie_value, 'encrypt' );
					setcookie( $cookie_name, $cookie_value, time() + 365 * DAY_IN_SECONDS, COOKIEPATH, COOKIE_DOMAIN );

					$this->put_to_cache( $ip_address, $cookie_visitor_name, $cookie_visitor_address );
				} else {
					$cookie_value = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_COOKIE[ $cookie_name ] ), FILTER_UNSAFE_RAW ) );
					$cookie_value = $this->encrypt_decrypt( $cookie_value, 'decrypt' );

					$cookie_array = explode( self::SEPERATOR, $cookie_value );

					$cookie_ip              = $ip_address;
					$cookie_time            = $cookie_array[1];
					$cookie_visitor_name    = $this->generate_visitorname( $cookie_array[2] );
					$cookie_visitor_address = $visitor_address_from_file;

					$cookie_value = $cookie_ip . self::SEPERATOR . $cookie_time . self::SEPERATOR . $cookie_visitor_name . self::SEPERATOR . $cookie_visitor_address;
					$cookie_value = $this->encrypt_decrypt( $cookie_value, 'encrypt' );
					setcookie( $cookie_name, $cookie_value, time() + 365 * DAY_IN_SECONDS, COOKIEPATH, COOKIE_DOMAIN );

					$this->put_to_cache( $ip_address, $cookie_visitor_name, $cookie_visitor_address );
				}
				$cookie_value = $this->encrypt_decrypt( $cookie_value, 'decrypt' );
			}
		}

		/**
		 * Returns API key printed in inputbox.
		 * Defaults to empty unless set.
		 *
		 * @return string $str Text.
		 */
		public function api_key() {
			$api_key = get_option( self::WPSA_API_KEY, '' );

			$str = "<input id='api_key' type='text' value='" . $api_key . "' />";

			return $str;
		}

		/**
		 * Returns comma seperated strings printed in textarea.
		 * Vistor paths containing substring of these strings are ignored.
		 *
		 * @return string $str Text.
		 */
		public function ignore_files() {
			$ignore_files_default = "'wp-cron', 'wp-json', 'robots.txt', 'security.txt', 'favicon.ico'";
			$ignore_files         = get_option( self::IGNORE_FILES, $ignore_files_default );

			$str = "<textarea id='ignore_files'  class='txtarea'>" . $ignore_files . '</textarea>';

			return $str;
		}

		/**
		 * Returns tag frequency printed in inputbox.
		 * Set to 3 during initialization.
		 *
		 * @return string $str Text.
		 */
		public function tag_frequency() {
			$tag_frequency = get_option( self::TAG_FREQUENCY );

			$str = "<input id='tag_frequency' type='text' value='" . $tag_frequency . "' />";

			return $str;
		}

		/**
		 * Returns session duration printed in inputbox.
		 * Set to 30 during initialization.
		 *
		 * @return string $str Text.
		 */
		public function session_duration() {
			$session_duration = get_option( self::SESSION_DURATION );

			$str = "<input id='session_duration' type='text' value='" . $session_duration . "' />";

			return $str;
		}

		/**
		 * Returns banned 'IP's text printed in textarea.
		 *
		 * @return string $str Text.
		 */
		public function meta_tags() {
			$tagarr = array();

			global $wp_filesystem;
			WP_Filesystem();
			$upload_dir    = wp_upload_dir();
			$wpsa_meta_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/meta';
			$path          = $wpsa_meta_dir . '/tags.txt';
			$str           = '';
			if ( file_exists( $path ) ) {
				$taglinestr = $wp_filesystem->get_contents( $path );
				$taglinestr = $this->encrypt_decrypt( $taglinestr, 'decrypt' );

				$str = "<textarea id='meta_tags' class='txtarea'>" . $taglinestr . '</textarea>';
			}
			return $str;
		}

		/**
		 * Returns toggle button for ignoring my visit using cookie.
		 * Defaults to false unless set.
		 *
		 * @return string $str Text.
		 */
		public function ignore_my_visits_cookie() {
			$set_cookie = '';

			$cookie_name = self::COOKIE_NAME;
			if ( isset( $_COOKIE[ $cookie_name ] ) && ! empty( $_COOKIE[ $cookie_name ] ) ) {
				$cookie_value = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_COOKIE[ $cookie_name ] ), FILTER_UNSAFE_RAW ) );
				$cookie_value = $this->encrypt_decrypt( $cookie_value, 'decrypt' );
				$cookie_array = explode( self::SEPERATOR, $cookie_value );
				if ( 4 === count( $cookie_array ) ) {
					$visitor_name = $cookie_array[2];
					if ( false !== stripos( $visitor_name, 'BlockMe' ) ) {
						$visitor_name = $cookie_array[2];
						$set_cookie   = "checked = 'checked'";
					}
				}
			}

			$str = "<label for='ignore_my_visits_cookie' class='switch'>";
			$str = $str . "<input type='checkbox' id='ignore_my_visits_cookie'" . $set_cookie . ' />';
			$str = $str . "<span class='ignore_my_visits_cookie slider round'></span>";
			$str = $str . '</label>';

			return $str;
		}

		/**
		 * Returns toggle button for ignoring my visit using IP.
		 * Defaults to false unless set.
		 *
		 * @return string $str Text.
		 */
		public function ignore_my_visits_ip() {
			$my_ip     = $this->get_ip_address( true );
			$ip_string = get_option( self::IP_STRING, '' );

			$set_ip = '';
			if ( stripos( $ip_string, $my_ip ) > 0 ) {
				$set_ip = " checked = 'checked'";
			}

			$str = "<label for='ignore_my_visits_ip' class='switch'>";
			$str = $str . "<input type='checkbox' id='ignore_my_visits_ip'" . $set_ip . ' />';
			$str = $str . "<span class='ignore_my_visits_ip slider round'></span>";
			$str = $str . '</label>';

			return $str;
		}

		/**
		 * Returns toggle button for ignoring bots.
		 * Defaults to false unless set.
		 *
		 * @return string $str Text.
		 */
		public function ignore_bots() {
			$ignore_bots = get_option( self::IGNORE_BOTS, false ) ? " checked = 'checked'" : '';

			$str = "<label for='ignore_bots' class='switch'>";
			$str = $str . "<input type='checkbox' id='ignore_bots'" . $ignore_bots . ' />';
			$str = $str . "<span class='ignore_bots slider round'></span>";
			$str = $str . '</label>';

			return $str;
		}

		/**
		 * Returns toggle button to anonymize ip.
		 * Defaults to true unless set.
		 *
		 * @return string $str Text.
		 */
		public function anonymize_ip() {
			$anonymize_ip = get_option( self::ANONYMIZE_IP, true ) ? " checked = 'checked'" : '';

			$str = "<label for='anonymize_ip' class='switch'>";
			$str = $str . "<input type='checkbox' id='anonymize_ip'" . $anonymize_ip . ' />';
			$str = $str . "<span class='anonymize_ip slider round'></span>";
			$str = $str . '</label>';

			return $str;
		}

		/**
		 * Returns toggle button to stop logging.
		 * Defaults to false unless set.
		 *
		 * @return string $str Text.
		 */
		public function stop_logging() {
			$stop_logging = get_option( self::STOP_LOGGING, false ) ? " checked = 'checked'" : '';

			$str = "<label for='stop_logging' class='switch'>";
			$str = $str . "<input type='checkbox' id='stop_logging'" . $stop_logging . ' />';
			$str = $str . "<span class='stop_logging slider round'></span>";
			$str = $str . '</label>';

			return $str;
		}

		/**
		 * Returns toggle button for reseting visitor count.
		 * Defaults to false always.
		 *
		 * @return string $str Text.
		 */
		public function reset_visitor_count() {
			$str = "<label for='reset_visitor_count' class='switch'>";
			$str = $str . "<input type='checkbox' id='reset_visitor_count' />";
			$str = $str . "<span class='reset_visitor_count slider round'></span>";
			$str = $str . '</label>';

			return $str;
		}

		/**
		 * Returns duration of today's log refresh printed in inputbox.
		 * Defaults to 1. In minutes. Allows decimal.
		 *
		 * @return string $str Text.
		 */
		public function refresh_todays_log_duration() {
			$refresh_todays_log_duration = get_option( self::REFRESH_TODAYS_LOG_DURATION, 1 );

			$str = "<input id='refresh_todays_log_duration' type='text' value='" . $refresh_todays_log_duration . "' />";

			return $str;
		}

		/**
		 * Returns maximum size of visitor cache records printed in inputbox.
		 * Set to 10000 during initialization.
		 *
		 * @return string $str Text.
		 */
		public function max_cache_records() {
			$max_cache_records = get_option( self::MAX_CACHE_RECORDS );

			$str = "<input id='max_cache_records' type='text' value='" . $max_cache_records . "' />";

			return $str;
		}

		/**
		 * Returns maximum size of visitor database records printed in inputbox.
		 * Set to 100000 during initialization.
		 * Immediate purge on save.
		 *
		 * @return string $str Text.
		 */
		public function max_database_records() {
			$max_database_records = get_option( self::MAX_DATABASE_RECORDS );

			$str = "<input id='max_database_records' type='text' value='" . $max_database_records . "' />";

			return $str;
		}

		/**
		 * Returns cloud tag and logs from server. If no filename is posted it takes today's log.
		 *
		 * @param string $access Access level (manage_options or read or shortcode).
		 * @param string $file Log file to display. Defaults to empty.
		 *
		 * @return string $str Text.
		 */
		public function view_logs( $access, $file = '' ) {
			$str = '';

			unset( $tag );
			$tag = array();
			global $wp_filesystem;
			WP_Filesystem();
			$upload_dir    = wp_upload_dir();
			$wpsa_meta_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/meta';
			$path          = $wpsa_meta_dir . '/tags.txt';
			if ( file_exists( $path ) ) {
				$taglinestr = $wp_filesystem->get_contents( $path );
				$taglinestr = $this->encrypt_decrypt( $taglinestr, 'decrypt' );

				$taglinearr = explode( PHP_EOL, $taglinestr );
				foreach ( $taglinearr as $tagline ) {
					$tagline = trim( $tagline );
					if ( '' !== $tagline ) {
						unset( $tagarr );
						$tagarr = array();
						$tagarr = explode( '=', $tagline );
						if ( isset( $tag[ "{$tagarr[0]}" ] ) ) {
							array_push( $tag[ "{$tagarr[0]}" ], $tagarr[1] );
						}
					}
				}
			}

			if ( empty( $file ) ) {
				$file = $this->get_log_filename();
			}

			$tagcloud     = array();
			$all_tagvalue = 0;

			global $wp_filesystem;
			WP_Filesystem();
			$upload_dir   = wp_upload_dir();
			$wpsa_log_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/log';
			$path         = $wpsa_log_dir . '/' . $file;
			if ( file_exists( $path ) ) {
				$content     = $wp_filesystem->get_contents( $path );
				$content_arr = explode( PHP_EOL, $content );
				foreach ( $content_arr as $content_line ) {
					$content_line = $this->encrypt_decrypt( $content_line, 'decrypt' ); // Log content is decrypted.

					unset( $tagarr );

					if ( 1 === preg_match( '/[A-Z]{1}[a-z\-]{1,}_\d{1}\d*/', $content_line, $matches ) ) {
						$visitor = $matches[0];
						unset( $matches );
					}

					// Check for IPv6 then IPv4 and then Loopback address.
					// Allows characater X also in the regex below.
					$ip_matched = '';
					$ip_version = '';
					if ( 1 === preg_match( '/(?:^|(?<=\s))(([0-9a-fA-FX]{1,4}:){7,7}[0-9a-fA-FX]{1,4}|([0-9a-fA-FX]{1,4}:){1,7}:|([0-9a-fA-FX]{1,4}:){1,6}:[0-9a-fA-FX]{1,4}|([0-9a-fA-FX]{1,4}:){1,5}(:[0-9a-fA-FX]{1,4}){1,2}|([0-9a-fA-FX]{1,4}:){1,4}(:[0-9a-fA-FX]{1,4}){1,3}|([0-9a-fA-FX]{1,4}:){1,3}(:[0-9a-fA-FX]{1,4}){1,4}|([0-9a-fA-FX]{1,4}:){1,2}(:[0-9a-fA-FX]{1,4}){1,5}|[0-9a-fA-FX]{1,4}:((:[0-9a-fA-FX]{1,4}){1,6})|:((:[0-9a-fA-FX]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-FX]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-FX]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))(?=\s|$)/i', $content_line, $matches ) ) {
						$ip_matched = $matches[0];
						$ip_version = 'IPv6';
					}
					if ( 1 === preg_match( '/[\dX]{1,3}\.[\dX]{1,3}\.[\dX]{1,3}\.[\dX]{1,3}/', $content_line, $matches ) ) {
						$ip_matched = $matches[0];
						$ip_version = 'IPv4';
					}
					if ( 1 === preg_match( '/0000:0000:0000:XXXX:XXXX:XXXX:XXXX:XXX/', $content_line, $matches ) ) {
						$ip_matched = $matches[0];
						$ip_version = 'Loopback address';
					}

					if ( '' !== $ip_matched ) {
						++$all_tagvalue;
						unset( $matches );

						if ( isset( $tag[ $ip_matched ] ) ) {

							$tagarr = array();
							$tagarr = $tag[ $ip_matched ];

							foreach ( $tagarr as $tagitem ) {
								if ( ! isset( $tagcloud[ "{$tagitem}" ] ) ) {
									$tagcloud[ "{$tagitem}" ] = 0;
								}
								++$tagcloud[ "{$tagitem}" ];
							}
						}

						if ( ! isset( $tagcloud[ "{$visitor}" ] ) ) {
							$tagcloud[ "{$visitor}" ] = 0;
						}
						++$tagcloud[ "{$visitor}" ];

						if ( ! isset( $tagcloud[ "{$ip_version}" ] ) ) {
							$tagcloud[ "{$ip_version}" ] = 0;
						}
						++$tagcloud[ "{$ip_version}" ];
					}
				}
			}

			arsort( $tagcloud ); // Sort by value with highest frequency first.
			$tagcloudcount = 0;
			$cloudstr      = "<div id='tagcloud'>";
			$cloudstr      = $cloudstr . "<span class='tag underline' onclick=\"view_tag_logs('all');\"  style='background-color:" . $this->get_lightcolor( 'all' ) . '; color:' . $this->get_darkcolor( 'all' ) . "'> all (" . $all_tagvalue . ')</span>';

			foreach ( $tagcloud as $tagname => $tagvalue ) {
				if ( $tagvalue >= get_option( self::TAG_FREQUENCY ) ) {  // How many user visits are required for a tag to be displayed?
					$tagcloudcount++;
					$cloudstr = $cloudstr . "<span class='tag underline' onclick=\"view_tag_logs('" . $this->sanitize_classname( $tagname ) . "');\" style='background-color:" . $this->get_lightcolor( $tagname ) . '; color:' . $this->get_darkcolor( $tagname ) . "'>" . $tagname . ' (' . $tagvalue . ')</span>';
				}
			}
			$cloudstr = $cloudstr . '</div>';

			if ( $tagcloudcount > 0 ) {
				$str = $str . $cloudstr;
			}

			$str = $str . "<div id='error_messages'></div>";

			global $wp_filesystem;
			WP_Filesystem();
			$upload_dir   = wp_upload_dir();
			$wpsa_log_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/log';
			$path         = $wpsa_log_dir . '/' . $file;
			if ( file_exists( $path ) ) {
				$content     = $wp_filesystem->get_contents( $path );
				$content_arr = explode( PHP_EOL, $content );
				$content_arr = explode( PHP_EOL, $content );
				foreach ( $content_arr as $content_line ) {
					$content_line = $this->encrypt_decrypt( $content_line, 'decrypt' ); // Log content is decrypted.

					unset( $matches );
					unset( $tagarr );
					$tagstr   = '';
					$classstr = 'all';

					if ( 1 === preg_match( '/[A-Z]{1}[a-z\-]{1,}_\d{1}\d*/', $content_line, $matches ) ) {
						$visitor = $matches[0];
					}

					// Check for IPv6 then IPv4 and then Loopback address.
					// Allows characater X also in the regex below.
					$ip_matched = '';
					$ip_version = '';
					if ( 1 === preg_match( '/(?:^|(?<=\s))(([0-9a-fA-FX]{1,4}:){7,7}[0-9a-fA-FX]{1,4}|([0-9a-fA-FX]{1,4}:){1,7}:|([0-9a-fA-FX]{1,4}:){1,6}:[0-9a-fA-FX]{1,4}|([0-9a-fA-FX]{1,4}:){1,5}(:[0-9a-fA-FX]{1,4}){1,2}|([0-9a-fA-FX]{1,4}:){1,4}(:[0-9a-fA-FX]{1,4}){1,3}|([0-9a-fA-FX]{1,4}:){1,3}(:[0-9a-fA-FX]{1,4}){1,4}|([0-9a-fA-FX]{1,4}:){1,2}(:[0-9a-fA-FX]{1,4}){1,5}|[0-9a-fA-FX]{1,4}:((:[0-9a-fA-FX]{1,4}){1,6})|:((:[0-9a-fA-FX]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-FX]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-FX]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))(?=\s|$)/i', $content_line, $matches ) ) {
						$ip_matched = $matches[0];
						$ip_version = 'IPv6';
					}
					if ( 1 === preg_match( '/[\dX]{1,3}\.[\dX]{1,3}\.[\dX]{1,3}\.[\dX]{1,3}/', $content_line, $matches ) ) {
						$ip_matched = $matches[0];
						$ip_version = 'IPv4';
					}
					if ( 1 === preg_match( '/0000:0000:0000:XXXX:XXXX:XXXX:XXXX:XXX/', $content_line, $matches ) ) {
						$ip_matched = $matches[0];
						$ip_version = 'Loopback address';
					}

					if ( '' !== $ip_matched ) {
						if ( isset( $tag[ $ip_matched ] ) ) {
							$tagarr = $tag[ $ip_matched ];
							foreach ( $tagarr as $tagitem ) {
								$tagstr   = $tagstr . "<span class='tag' style='background-color:" . $this->get_lightcolor( $tagitem ) . '; color:' . $this->get_darkcolor( $tagitem ) . "'>" . $tagitem . '</span>';
								$classstr = $classstr . ' ' . $this->sanitize_classname( $tagitem );
							}
						}

						if ( isset( $ip_matched ) ) {
							$tagitem  = $visitor;
							$tagstr   = $tagstr . "<span class='tag' style='background-color:" . $this->get_lightcolor( $tagitem ) . '; color:' . $this->get_darkcolor( $tagitem ) . "'>" . $tagitem . '</span>';
							$classstr = $classstr . ' ' . $this->sanitize_classname( $tagitem );
						}

						if ( isset( $ip_version ) ) {
							$tagitem  = $ip_version;
							$tagstr   = $tagstr . "<span class='tag' style='background-color:" . $this->get_lightcolor( $tagitem ) . '; color:' . $this->get_darkcolor( $tagitem ) . "'>" . $tagitem . '</span>';
							$classstr = $classstr . ' ' . $this->sanitize_classname( $tagitem );
						}
					}

					if ( false !== strpos( $content_line, '.XXX' ) ) {
						$content_line = preg_replace( '/<span([^>]+)\>|<\/span>/i', "<span class='readtaglink'>", $content_line, 1 );
					}

					if ( ( 'read' === $access ) || ( 'shortcode' === $access ) ) {
						$content_line = preg_replace( '/<span([^>]+)\>|<\/span>/i', "<span class='readtaglink'>", $content_line, 1 );
					}

					$content_line = str_replace( 'table_row', 'table_row ' . $classstr, $content_line );
					$content_line = str_replace( '</div>', $tagstr . '</div>', $content_line );

					$str = $str . $content_line;
				}
			}

			if ( $file === $this->get_log_filename() ) { // Viewing todays logs?
				$str = $str . "<script type='text/javascript'>window.clearInterval(timer); timer =window.setInterval(function(){ajax_view_log('" . $file . "');}, Math.round(" . get_option( self::REFRESH_TODAYS_LOG_DURATION, 1 ) . '* 60* 1000));</script>';
			} else {
				$str = $str . "<script type='text/javascript'>window.clearInterval(timer);</script>";
			}

			return $str;
		}

		/**
		 * Get the current log file.
		 *
		 * @return string $log_filename Today's log filename.
		 */
		public function get_log_filename() {
			$timestamp    = time();
			$log_filename = gmdate( 'D_d_M_Y', $timestamp );
			$log_filename = $log_filename . '.txt';

			return $log_filename;
		}
		/**
		 * --------------------------------Frontend Section-----------------------------------------.
		 */
		/**
		 * Retrieves address from File 'cache.txt' if available. False if it is not available.
		 *
		 * @param string $ip_address IP address.
		 *
		 * @return array $line_data Cached information.
		 */
		public function get_from_cache( $ip_address ) {
			global $wp_filesystem;
			WP_Filesystem();
			$upload_dir    = wp_upload_dir();
			$wpsa_meta_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/meta';
			$path          = $wpsa_meta_dir . '/cache.txt';

			if ( file_exists( $path ) ) {
				$file_data = $wp_filesystem->get_contents( $path );
				if ( $file_data ) {
					$file_data_array = explode( PHP_EOL, $file_data );
					foreach ( $file_data_array as $line_data ) {
						$line_data       = $this->encrypt_decrypt( $line_data, 'decrypt' );
						$line_data_array = explode( self::SEPERATOR, $line_data );
						if ( 4 === count( $line_data_array ) ) {
							$ip_address_temp = $line_data_array[0];
							$address_temp    = $line_data_array[3];
							if ( trim( $ip_address_temp ) === trim( $ip_address ) ) {
								return $line_data_array;
							}
						}
					}
				}
			}

			return false;
		}

		/**
		 * Writes address corresponding to File 'cache.txt' if available.
		 *
		 * @param string $ip_address IP address.
		 * @param string $visitor_name Visitor name allocated.
		 * @param string $address Address.
		 *
		 * @return void
		 */
		public function put_to_cache( $ip_address, $visitor_name, $address ) {
			global $wp_filesystem;
			WP_Filesystem();
			$upload_dir    = wp_upload_dir();
			$wpsa_meta_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/meta';
			$path          = $wpsa_meta_dir . '/cache.txt';
			$file_data     = '';

			$max_lines = get_option( self::MAX_CACHE_RECORDS );
			if ( $max_lines > 0 ) {
				$new_file_data = $ip_address . self::SEPERATOR . time() . self::SEPERATOR . $visitor_name . self::SEPERATOR . $address;
				$new_file_data = $this->encrypt_decrypt( $new_file_data, 'encrypt' );

				if ( file_exists( $path ) ) {
					$file_data = $wp_filesystem->get_contents( $path );
				}

				$lines           = 0;
				$file_data_array = explode( PHP_EOL, $file_data );
				foreach ( $file_data_array as $line_data ) {
					$lines++;
					if ( $lines < $max_lines ) { // < Since we have a new record to add.
						$new_file_data = $new_file_data . PHP_EOL . $line_data;
					}
				}

				$wp_filesystem->put_contents( $path, $new_file_data, 0644 );
			} else {
				if ( file_exists( $path ) ) {
					$wp_filesystem->put_contents( $path, '', 0644 );
				}
			}
		}

		/**
		 * Logs page load and writes the log information to todays log text file in 'logs' directory.
		 *
		 * @return void
		 */
		public function log_visitor() {
			global $wpdb;

			$ip_address = $this->get_ip_address( true );
			$ip_version = '';
			if ( false !== filter_var( $ip_address, FILTER_VALIDATE_IP, FILTER_FLAG_IPV4 ) ) {
				$ip_version = 'IPv4';
			} elseif ( false !== filter_var( $ip_address, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6 ) ) {
				$ip_version = 'IPv6'; // Loopback address is special case of IPv6.
			} else {
				return; // Ignore invalid IP.
			}

			if ( $this->check_ajax() ) {
				return; // Ignore AJAX calls.
			}

			$user_agent = '';
			if ( isset( $_SERVER['HTTP_USER_AGENT'] ) && ( ! empty( $_SERVER['HTTP_USER_AGENT'] ) ) ) {
				$user_agent = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_SERVER['HTTP_USER_AGENT'] ), FILTER_UNSAFE_RAW ) );
			}
			$bot_info        = $this->bot_info( $user_agent );
			$platform        = $this->os_info( $user_agent );
			$browser_name    = $this->browser_info( $user_agent, 'browser' );
			$browser_version = $this->browser_info( $user_agent, 'version' );

			$replace_name = '';
			$bot_name     = '';

			if ( false !== $bot_info ) {
				$bot_name     = $bot_info;
				$replace_name = ucwords( strtolower( str_replace( ' ', '-', $bot_info ) . '-bot' ) ); // We have captured bot name to tag the visit.
				if ( get_option( self::IGNORE_BOTS, false ) ) {
					return; // Ignore bot calls. No need to log them.
				}
			}

			$ignore_files_default = "'wp-cron', 'wp-json', 'robots.txt', 'security.txt', 'favicon.ico'";
			$filter               = get_option( self::IGNORE_FILES, $ignore_files_default );
			$filter               = array_map( 'trim', explode( ',', $filter ) );

			if ( $this->check_substring( $this->get_url(), $filter ) ) {
				return; // Do not log (File paths containing any one of the comma seperated strings below are ignored).
			}

			$session_name = self::COOKIE_NAME;
			if ( isset( $_SESSION[ $session_name ] ) && ! empty( $_SESSION[ $session_name ] ) ) {
				$session_value = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_SESSION[ $session_name ] ), FILTER_UNSAFE_RAW ) );
				$session_value = $this->encrypt_decrypt( $session_value, 'decrypt' );
				$session_array = explode( self::SEPERATOR, $session_value );
				if ( 4 === count( $session_array ) ) {
					$visitor_name = $session_array[2];
					if ( false !== stripos( $visitor_name, 'BlockMe' ) ) {
						return; // Ignore my visits (Using cookies).
					}
				}
			}

			$my_ip     = $this->get_ip_address( true );
			$ip_string = get_option( self::IP_STRING, '' );

			$set_ip = '';
			if ( stripos( $ip_string, $my_ip ) > 0 ) {
				return; // Ignore my visits (Using IP).
			}

			$ip_address = $this->get_ip_address();
			$time       = time();
			$duration   = 0;

			$session_name = self::COOKIE_NAME;
			if ( isset( $_SESSION[ $session_name ] ) && ! empty( $_SESSION[ $session_name ] ) ) {
				$session_value = $_SESSION[ $session_name ];

				$session_value = $this->encrypt_decrypt( $session_value, 'decrypt' );

				$session_array = explode( self::SEPERATOR, $session_value );

				$session_previous_time = $session_array[1];
				$visitor_name          = $session_array[2];
				$visitor_address       = $session_array[3];

				if ( $session_previous_time > 0 ) {
					$duration = $time - $session_previous_time; // Calculates how much time the visitor stayed on previous page.
				}
			} else {

				$time_from_file            = 0;
				$visitor_name_from_file    = '';
				$visitor_address_from_file = '';
				if ( '' !== $ip_address ) {
					$cache_from_file = $this->get_from_cache( $ip_address );

					if ( false === $cache_from_file ) {
						$ip_location               = $this->get_ip_location_details( $ip_address );
						$visitor_address_from_file = $this->get_visitor_location_from_ip_location_details( $ip_location );
					} else {
						$time_from_file            = $cache_from_file[1];
						$visitor_name_from_file    = $cache_from_file[2];
						$visitor_address_from_file = $cache_from_file[3];
					}

					if ( $time_from_file > 0 ) {
						$duration = $time - $time_from_file; // Calculates how much time the visitor stayed on previous page.
					}
				}

				$visitor_name    = $visitor_name_from_file;
				$visitor_address = $visitor_address_from_file;
			}

			$visitor_name = $this->generate_visitorname( $visitor_name, $replace_name );
			$page_type    = $this->get_page_type();

			$referrer_url = '';
			if ( isset( $_SERVER['HTTP_REFERER'] ) && ( ! empty( $_SERVER['HTTP_REFERER'] ) ) ) {
				$referrer_url = filter_var( wp_unslash( $_SERVER['HTTP_REFERER'] ), FILTER_VALIDATE_URL );
			}
			$referrer_url = $this->get_referrer( $referrer_url );

			$accept_languages = '';
			if ( isset( $_SERVER['HTTP_ACCEPT_LANGUAGE'] ) && ( ! empty( $_SERVER['HTTP_ACCEPT_LANGUAGE'] ) ) ) {
				$accept_languages = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_SERVER['HTTP_ACCEPT_LANGUAGE'] ), FILTER_UNSAFE_RAW ) );
			}
			$accept_languages = $this->get_user_base_language( $accept_languages, self::ACCEPT_LANGUAGE_PRIORITY, self::ACCEPT_LANGUAGE_LIMIT );

			$title        = $this->get_title();
			$url          = $this->get_url();
			$current_time = $this->get_current_time();

			$str = "<div class='table_row' style='border-color:" . $this->get_darkcolor( $visitor_name ) . "'>";

			$str = $str . "<span class='taglink' onclick=\"add_tag('" . $ip_address . "')\">" . $visitor_name . '</span><b>' . $current_time . ' : </b>';
			$str = $str . __( 'A visitor with ', 'wpsa' ) . $ip_version . ' ' . $ip_address . __( ' has visited the ', 'wpsa' );

			if ( '' === $title ) {
				$str = $str . "<a target='_blank' href='" . $url . "'>" . ( '' === $page_type ? 'website. ' : $page_type . '. ' ) . '</a>';
			} else {
				$str = $str . ( '' === $page_type ? 'website' : $page_type ) . __( ' with title ', 'wpsa' ) . "<a target='_blank' href='" . $url . "'>" . $title . '</a>. ';
			}

			if ( strlen( $visitor_address ) > 0 ) {
				$str = $str . __( 'Visitor location is ', 'wpsa' ) . "<span style='color:" . $this->get_darkcolor( $visitor_address ) . "'>" . $visitor_address . '. </span>';
			}

			if ( false === $bot_info ) {
				$str = $str . __( 'Visitor is using ', 'wpsa' ) . "<span style='color:" . $this->get_darkcolor( $browser_name ) . "'>" . $browser_name . __( ' browser', 'wpsa' ) . ( 'Unknown' !== $browser_version ? __( ' of version ', 'wpsa' ) . $browser_version : '' ) . '</span>' . __( ' on ', 'wpsa' ) . "<span style='color:" . $this->get_darkcolor( $platform ) . "'>" . $platform . __( ' OS', 'wpsa' ) . '. </span>';
			} else {
				$str = $str . __( 'Visitor is ', 'wpsa' ) . "<span style='color:" . $this->get_darkcolor( $replace_name ) . "'>" . $bot_name . ' Bot. </span>';
			}

			if ( '' !== $accept_languages ) {
				$str = $str . __( 'Visitor prefers ', 'wpsa' ) . $accept_languages . '. ';
			}

			$temp_referrer_url = '';
			if ( isset( $_SERVER['HTTP_REFERER'] ) && ( ! empty( $_SERVER['HTTP_REFERER'] ) ) ) {
				$temp_referrer_url = filter_var( wp_unslash( $_SERVER['HTTP_REFERER'] ), FILTER_VALIDATE_URL );
			}
			$referrer_source = $this->get_referrer_source( $temp_referrer_url );

			if ( '' !== $referrer_url ) {
				$str = $str . __( 'Visitor has been referred from link at ', 'wpsa' ) . "<a target='_blank' href='" . $referrer_url . "'>" . $referrer_source . '</a>.';
			} else {
				$str = $str . '<span class="time_placeholder">' . $visitor_name . '</span>';
			}
			$str = $str . '</div>';

			$str = $this->encrypt_decrypt( $str, 'encrypt' ); // Log content is encrypted.

			global $wp_filesystem;
			WP_Filesystem();
			$upload_dir   = wp_upload_dir();
			$wpsa_log_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/log';
			$path         = $wpsa_log_dir . '/' . $this->get_log_filename();
			$file_data    = '';
			if ( file_exists( $path ) ) {
				$file_data = $wp_filesystem->get_contents( $path );
			}

			$search_data  = '<span class="time_placeholder">' . $visitor_name . '</span>';
			$replace_data = '';
			if ( 0 === $duration ) {
				$replace_data = __( ' Visitor has left page immediately.', 'wpsa' );
			} elseif ( $duration <= get_option( self::SESSION_DURATION ) * 60 ) { // Multiplied by 60 to convert to minutes.
				$replace_data = __( ' Visitor has spent ', 'wpsa' ) . $this->duration_verbose( $duration ) . __( ' on the page.', 'wpsa' );
			} elseif ( $duration > get_option( self::SESSION_DURATION ) * 60 ) {
				$replace_data = __( ' Visitor may have left the session as the visitor has spend more than ', 'wpsa' ) . $this->duration_verbose( $duration ) . __( ' on the page.', 'wpsa' );
			}

			$file_data_arr = explode( PHP_EOL, $file_data );
			foreach ( $file_data_arr as $key => $file_data_item ) {
				$file_data_item = $this->encrypt_decrypt( $file_data_item, 'decrypt' ); // Log content is read and decrypted one line at a time.
				$pos            = stripos( $file_data_item, $search_data );
				if ( false !== $pos ) {
					$file_data_item        = substr_replace( $file_data_item, $replace_data, $pos, strlen( $search_data ) );
					$file_data_item        = $this->encrypt_decrypt( $file_data_item, 'encrypt' ); // Required log content is updated and ecrypted.
					$file_data_arr[ $key ] = $file_data_item;
					break;
				}
			}
			$file_data = implode( PHP_EOL, $file_data_arr );

			if ( file_exists( $path ) ) {
				$file_data = $str . PHP_EOL . $file_data;
			} else {
				$file_data = $str;
			}

			if ( ! get_option( self::STOP_LOGGING, false ) ) {
				$wp_filesystem->put_contents( $path, $file_data, 0644 ); // No need to log them.
			}

			$page_type = ucwords( $page_type );
			if ( '' === $title ) {
				$title = $page_type;
			}

			$country_array = explode( ',', $visitor_address );
			$country       = array_pop( $country_array );
			if ( is_null( $country ) ) {
				$country = '';
			}

			$is_bot = true;
			if ( false === stripos( $visitor_name, 'bot_' ) ) {
				$is_bot = false;
			}

			$is_admin = true;
			if ( false === stripos( $visitor_name, 'admin_' ) ) {
				$is_admin = false;
			}

			$table_name = $wpdb->prefix . 'smart_analytics_visitor_log';

			$max_database_records = get_option( self::MAX_DATABASE_RECORDS, 100000 );

			if ( $max_database_records > 0 ) {
				$this->wp_smart_analytics_create_plugin_database_table();

				if ( ( $duration <= get_option( self::SESSION_DURATION ) * 60 ) && ( $duration > 0 ) ) { // Multiplied by 60 to convert to minutes. Valid duration.
					$wpdb->query( $wpdb->prepare( "UPDATE $table_name SET time_in_sec = %d WHERE name = %s ORDER BY time DESC LIMIT 1", $duration, $visitor_name ) );
				}

				$data   = array(
					'ip'               => $ip_address,
					'ip_version'       => $ip_version,
					'url'              => $url,
					'page_type'        => $page_type,
					'title'            => $title,
					'name'             => $visitor_name,
					'address'          => $visitor_address,
					'country'          => $country,
					'bot_info'         => $bot_info,
					'platform'         => $platform,
					'browser_name'     => $browser_name,
					'browser_version'  => $browser_version,
					'accept_languages' => $accept_languages,
					'referrer_url'     => $referrer_url,
					'referrer_source'  => $referrer_source,
					'is_bot'           => $is_bot,
					'is_admin'         => $is_admin,
					'time_in_sec'      => 0,
				);
				$format = array( '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%d', '%d', '%d' );
				$wpdb->insert( $table_name, $data, $format );
			}

			$wpdb->query( $wpdb->prepare( "DELETE FROM $table_name WHERE id IN ( SELECT id FROM (SELECT id FROM wp_smart_analytics_visitor_log ORDER BY id DESC LIMIT %d, 2147483647) x);", $max_database_records ) );
		}

		/**
		 * --------------------------------AJAX Section-----------------------------------------.
		 */
		/**
		 * Calls view_logs function. This is a Ajax specific call which requires exit (ajax).
		 *
		 * @return void
		 */
		public function wp_ajax_wpsa_view_log() {
			if ( isset( $_POST ) ) {
				$_POST = filter_input_array( INPUT_POST, FILTER_UNSAFE_RAW );
			}

			if ( wp_verify_nonce( isset( $_POST['nonce'] ) ? filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['nonce'] ), FILTER_UNSAFE_RAW ) ) : '', 'wpsa_nonce' ) ) {
				if ( isset( $_POST['file'] ) ) {
					$file = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['file'] ), FILTER_UNSAFE_RAW ) );
				}

				if ( current_user_can( 'manage_options' ) ) {
					echo $this->view_logs( 'manage_options', $file );
				} elseif ( current_user_can( 'read' ) ) {
					echo $this->view_logs( 'read', $file );
				}
			}

			wp_die(); // Plugin is required to terminate immediately and return a proper response.
		}

		/**
		 * Delete the log from server (ajax).
		 *
		 * @return void
		 */
		public function wp_ajax_wpsa_delete_log() {
			if ( isset( $_POST ) ) {
				$_POST = filter_input_array( INPUT_POST, FILTER_UNSAFE_RAW );
			}

			if ( wp_verify_nonce( isset( $_POST['nonce'] ) ? filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['nonce'] ), FILTER_UNSAFE_RAW ) ) : '', 'wpsa_nonce' ) ) {
				if ( current_user_can( 'manage_options' ) ) {
					if ( isset( $_POST['file'] ) ) {
						$file = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['file'] ), FILTER_UNSAFE_RAW ) );

						$upload_dir   = wp_upload_dir();
						$wpsa_log_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/log/';
						wp_delete_file( $wpsa_log_dir . $file );

						$str = "<div class='message' style='border-color:#00a32a'>";
						$str = $str . __( 'The log file has been deleted successfully.', 'wpsa' );
						$str = $str . '</div>';

						echo $str;
					}
				}
			}

			wp_die(); // Plugin is required to terminate immediately and return a proper response.
		}

		/**
		 * Add tag (ip, tagname) in 'tags.txt' file in 'meta' folder (ajax).
		 *
		 * @return void
		 */
		public function wp_ajax_wpsa_add_tag() {
			if ( isset( $_POST ) ) {
				$_POST = filter_input_array( INPUT_POST, FILTER_UNSAFE_RAW );
			}

			if ( wp_verify_nonce( isset( $_POST['nonce'] ) ? filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['nonce'] ), FILTER_UNSAFE_RAW ) ) : '', 'wpsa_nonce' ) ) {
				if ( current_user_can( 'manage_options' ) ) {
					if ( isset( $_POST['ip'] ) && isset( $_POST['tag'] ) ) {
						$ip  = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['ip'] ), FILTER_UNSAFE_RAW ) );
						$tag = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['tag'] ), FILTER_UNSAFE_RAW ) );

						global $wp_filesystem;
						WP_Filesystem();
						$upload_dir    = wp_upload_dir();
						$wpsa_meta_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/meta';
						$path          = $wpsa_meta_dir . '/tags.txt';

						$str = "{$ip}={$tag}";
						if ( file_exists( $path ) ) {
							$file_data = $wp_filesystem->get_contents( $path );
							$file_data = $this->encrypt_decrypt( $file_data, 'decrypt' );
							if ( ! empty( $file_data ) ) {
								$file_data = trim( $file_data, PHP_EOL ) . PHP_EOL . $str;
							} else {
								$file_data = $str;
							}
						} else {
							$file_data = $str;
						}

						$file_data = $this->encrypt_decrypt( $file_data, 'encrypt' );
						$wp_filesystem->put_contents( $path, $file_data, 0644 );

						$str = "<div class='message' style='border-color:#00a32a'>";
						$str = $str . __( 'The ip : ', 'wpsa' ) . $ip . __( ' has been tagged with name : ', 'wpsa' ) . $tag . __( '. The screen will refresh in 3 seconds.', 'wpsa' );
						$str = $str . '</div>';

						echo $str;
					}
				}
			}

			wp_die(); // Plugin is required to terminate immediately and return a proper response.
		}

		/**
		 * Save settings.
		 *
		 * @return void
		 */
		public function wp_ajax_wpsa_save_settings() {
			global $wpdb;

			if ( isset( $_POST ) ) {
				$_POST = filter_input_array( INPUT_POST, FILTER_UNSAFE_RAW );
			}

			if ( wp_verify_nonce( isset( $_POST['nonce'] ) ? filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['nonce'] ), FILTER_UNSAFE_RAW ) ) : '', 'wpsa_nonce' ) ) {
				if ( current_user_can( 'manage_options' ) ) {
					if ( isset( $_POST['api_key'] ) ) {
						$api_key = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['api_key'] ), FILTER_UNSAFE_RAW ) );

						update_option( self::WPSA_API_KEY, $api_key );
					}

					if ( isset( $_POST['ignore_files'] ) ) {
						$ignore_files = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['ignore_files'] ), FILTER_UNSAFE_RAW ) );

						update_option( self::IGNORE_FILES, $ignore_files );
					}

					if ( isset( $_POST['tag_frequency'] ) ) {
						$tag_frequency = filter_var( wp_unslash( $_POST['tag_frequency'] ), FILTER_SANITIZE_NUMBER_INT );

						$tag_frequency = intval( $tag_frequency );
						if ( $tag_frequency > 0 ) {
							update_option( self::TAG_FREQUENCY, $tag_frequency );
						}
					}

					if ( isset( $_POST['session_duration'] ) ) {
						$session_duration = filter_var( wp_unslash( $_POST['session_duration'] ), FILTER_SANITIZE_NUMBER_INT );

						$session_duration = intval( $session_duration );
						if ( $session_duration > 0 ) {
							update_option( self::SESSION_DURATION, $session_duration );
						}
					}

					if ( isset( $_POST['meta_tags'] ) ) {
						$meta_tags = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['meta_tags'] ), FILTER_UNSAFE_RAW ) );

						global $wp_filesystem;
						WP_Filesystem();
						$upload_dir    = wp_upload_dir();
						$wpsa_meta_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/meta';
						$path          = $wpsa_meta_dir . '/tags.txt';
						$meta_tags     = $this->encrypt_decrypt( $meta_tags, 'encrypt' );
						$wp_filesystem->put_contents( $path, $meta_tags, 0644 );
					}

					if ( isset( $_POST['ignore_my_visits_cookie'] ) ) {
						$ignore_my_visits_cookie = filter_var( wp_unslash( $_POST['ignore_my_visits_cookie'] ), FILTER_VALIDATE_BOOLEAN );

						$ip_address = $this->get_ip_address();

						$time_from_file            = 0;
						$visitor_name_from_file    = '';
						$visitor_address_from_file = '';
						if ( '' !== $ip_address ) {
							$cache_from_file = $this->get_from_cache( $ip_address );

							if ( false === $cache_from_file ) {
								$ip_location               = $this->get_ip_location_details( $ip_address );
								$visitor_address_from_file = $this->get_visitor_location_from_ip_location_details( $ip_location );
							} else {
								$time_from_file            = $cache_from_file[1];
								$visitor_name_from_file    = $cache_from_file[2];
								$visitor_address_from_file = $cache_from_file[3];
							}
						}

						$cookie_name = self::COOKIE_NAME;

						$cookie_invalid = true; // Initially assuming cookie is invalid.
						if ( isset( $_COOKIE[ $cookie_name ] ) && ! empty( $_COOKIE[ $cookie_name ] ) ) { // Equivalent to else case below.
							$cookie_value = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_COOKIE[ $cookie_name ] ), FILTER_UNSAFE_RAW ) );
							$cookie_value = $this->encrypt_decrypt( $cookie_value, 'decrypt' );
							if ( 3 === substr_count( $cookie_value, self::SEPERATOR ) ) {
								$cookie_invalid = false;
							}
						}

						if ( ! isset( $_COOKIE[ $cookie_name ] ) || empty( $_COOKIE[ $cookie_name ] ) || $cookie_invalid ) {
							$cookie_ip              = $ip_address;
							$cookie_time            = 0;
							$visitor_name           = $visitor_name_from_file;
							$cookie_visitor_address = $visitor_address_from_file;
						} else {
							$cookie_value = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_COOKIE[ $cookie_name ] ), FILTER_UNSAFE_RAW ) );
							$cookie_value = $this->encrypt_decrypt( $cookie_value, 'decrypt' );

							$cookie_array = explode( self::SEPERATOR, $cookie_value );

							$cookie_ip              = $ip_address;
							$cookie_time            = $cookie_array[1];
							$visitor_name           = $cookie_array[2];
							$cookie_visitor_address = $cookie_array[3];
						}

						if ( true === $ignore_my_visits_cookie ) {
							$cookie_visitor_name = $this->generate_visitorname( $visitor_name, 'BlockMe' );

							$cookie_value = $cookie_ip . self::SEPERATOR . $cookie_time . self::SEPERATOR . $cookie_visitor_name . self::SEPERATOR . $cookie_visitor_address;
							$cookie_value = $this->encrypt_decrypt( $cookie_value, 'encrypt' );
							setcookie( $cookie_name, $cookie_value, time() + 365 * DAY_IN_SECONDS, COOKIEPATH, COOKIE_DOMAIN );
						} else {
							$cookie_visitor_name = $this->generate_visitorname( $visitor_name, 'Admin' );

							$cookie_value = $cookie_ip . self::SEPERATOR . $cookie_time . self::SEPERATOR . $cookie_visitor_name . self::SEPERATOR . $cookie_visitor_address;
							$cookie_value = $this->encrypt_decrypt( $cookie_value, 'encrypt' );
							setcookie( $cookie_name, $cookie_value, time() + 365 * DAY_IN_SECONDS, COOKIEPATH, COOKIE_DOMAIN );
						}

						$this->put_to_cache( $ip_address, $cookie_visitor_name, $cookie_visitor_address );
					}

					if ( isset( $_POST['ignore_my_visits_ip'] ) ) {
						$ignore_my_visits_ip = filter_var( wp_unslash( $_POST['ignore_my_visits_ip'] ), FILTER_VALIDATE_BOOLEAN );

						$ip_string     = get_option( self::IP_STRING, '' );
						$my_ip_address = $this->get_ip_address( true );
						if ( true === $ignore_my_visits_ip ) {
							$pos = strpos( $ip_string, $my_ip_address );

							if ( false === $pos ) {
								$ip_string = trim( $ip_string, '|' ) . '|' . $my_ip_address;
								update_option( self::IP_STRING, $ip_string );
							}
						} else {
							$ip_string = trim( str_replace( $my_ip_address, '', $ip_string ), '|' );
							update_option( self::IP_STRING, $ip_string );
						}
					}

					if ( isset( $_POST['ignore_bots'] ) ) {
						$ignore_bots = filter_var( wp_unslash( $_POST['ignore_bots'] ), FILTER_VALIDATE_BOOLEAN );

						update_option( self::IGNORE_BOTS, $ignore_bots );
					}

					if ( isset( $_POST['anonymize_ip'] ) ) {
						$anonymize_ip = filter_var( wp_unslash( $_POST['anonymize_ip'] ), FILTER_VALIDATE_BOOLEAN );

						update_option( self::ANONYMIZE_IP, $anonymize_ip );
					}

					if ( isset( $_POST['stop_logging'] ) ) {
						$stop_logging = filter_var( wp_unslash( $_POST['stop_logging'] ), FILTER_VALIDATE_BOOLEAN );

						update_option( self::STOP_LOGGING, $stop_logging );
					}

					if ( isset( $_POST['reset_visitor_count'] ) ) {
						$reset_visitor_count = filter_var( wp_unslash( $_POST['reset_visitor_count'] ), FILTER_VALIDATE_BOOLEAN );

						if ( true === $reset_visitor_count ) {
							update_option( self::VISITOR_COUNT, 1 );

							$upload_dir    = wp_upload_dir();
							$wpsa_meta_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/meta';
							$path          = $wpsa_meta_dir . '/cache.txt';
							if ( file_exists( $path ) ) {
								wp_delete_file( $path );
							}

							$ip_address = $this->get_ip_address();

							$visitor_name_from_file    = '';
							$visitor_address_from_file = '';
							if ( '' !== $ip_address ) {
								$cache_from_file = $this->get_from_cache( $ip_address );

								if ( false === $cache_from_file ) {
									$ip_location               = $this->get_ip_location_details( $ip_address );
									$visitor_address_from_file = $this->get_visitor_location_from_ip_location_details( $ip_location );
								}
							}

							$cookie_ip              = $ip_address;
							$cookie_time            = 0;
							$cookie_visitor_name    = $this->generate_visitorname( '', 'Admin' );
							$cookie_visitor_address = $visitor_address_from_file;

							$cookie_value = $cookie_ip . self::SEPERATOR . $cookie_time . self::SEPERATOR . $cookie_visitor_name . self::SEPERATOR . $cookie_visitor_address;
							$cookie_value = $this->encrypt_decrypt( $cookie_value, 'encrypt' );
							setcookie( $cookie_name, $cookie_value, time() + 365 * DAY_IN_SECONDS, COOKIEPATH, COOKIE_DOMAIN );

							$this->put_to_cache( $ip_address, $cookie_visitor_name, $cookie_visitor_address );
						}
					}

					if ( isset( $_POST['refresh_todays_log_duration'] ) ) {
						$refresh_todays_log_duration = filter_var( wp_unslash( $_POST['refresh_todays_log_duration'] ), FILTER_SANITIZE_NUMBER_INT );

						update_option( self::REFRESH_TODAYS_LOG_DURATION, $refresh_todays_log_duration );
					}

					if ( isset( $_POST['max_cache_records'] ) ) {
						$max_cache_records = filter_var( wp_unslash( $_POST['max_cache_records'] ), FILTER_SANITIZE_NUMBER_INT );

						update_option( self::MAX_CACHE_RECORDS, $max_cache_records );
					}

					if ( isset( $_POST['max_database_records'] ) ) {
						$max_database_records = filter_var( wp_unslash( $_POST['max_database_records'] ), FILTER_SANITIZE_NUMBER_INT );

						update_option( self::MAX_DATABASE_RECORDS, $max_database_records );
						if ( $max_database_records > 0 ) {
							$this->wp_smart_analytics_create_plugin_database_table();
						}
						$table_name = $wpdb->prefix . 'smart_analytics_visitor_log';
						$wpdb->query( $wpdb->prepare( "DELETE FROM $table_name WHERE id IN ( SELECT id FROM (SELECT id FROM wp_smart_analytics_visitor_log ORDER BY id DESC LIMIT %d, 2147483647) x);", $max_database_records ) );
					}
				}
			}

			wp_die(); // Plugin is required to terminate immediately and return a proper response.
		}

		/**
		 * Generate report.
		 * Output report rows as $str.
		 *
		 * @return void
		 */
		public function wp_ajax_wpsa_generate_report() {
			global $wpdb;

			$str = '';

			if ( isset( $_POST ) ) {
				$_POST = filter_input_array( INPUT_POST, FILTER_UNSAFE_RAW );
			}

			if ( wp_verify_nonce( isset( $_POST['nonce'] ) ? filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['nonce'] ), FILTER_UNSAFE_RAW ) ) : '', 'wpsa_nonce' ) ) {
				if ( current_user_can( 'manage_options' ) ) {
					$start_date          = '';
					$end_date            = '';
					$category_group      = '';
					$remove_bot_visits   = '';
					$remove_admin_visits = '';
					$sort                = 'category_group';
					$direction           = 'ASC';

					if ( isset( $_POST['start_date'] ) ) {
						$start_date = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['start_date'] ), FILTER_UNSAFE_RAW ) );
					}

					if ( isset( $_POST['end_date'] ) ) {
						$end_date = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['end_date'] ), FILTER_UNSAFE_RAW ) );
						$end_date = date_create( $end_date );
						date_add( $end_date, date_interval_create_from_date_string( '1 day' ) );
						$end_date = date_format( $end_date, 'Y-m-d' );
					}

					if ( isset( $_POST['category_group'] ) ) {
						$category_group = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['category_group'] ), FILTER_UNSAFE_RAW ) );
					}

					if ( isset( $_POST['remove_bot_visits'] ) ) {
						$remove_bot_visits     = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['remove_bot_visits'] ), FILTER_UNSAFE_RAW ) );
						$remove_bot_visits_str = 'true' === $remove_bot_visits ? ' AND is_bot != 1' : '';
					}

					if ( isset( $_POST['remove_admin_visits'] ) ) {
						$remove_admin_visits     = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['remove_admin_visits'] ), FILTER_UNSAFE_RAW ) );
						$remove_admin_visits_str = 'true' === $remove_admin_visits ? ' AND is_admin != 1' : '';
					}

					if ( isset( $_POST['sort'] ) ) {
						$sort = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['sort'] ), FILTER_UNSAFE_RAW ) );
					}

					if ( isset( $_POST['direction'] ) ) {
						$direction = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['direction'] ), FILTER_UNSAFE_RAW ) );
					}

					if ( ( gmdate( 'Y-m-d', strtotime( $start_date ) ) === $start_date ) && ( gmdate( 'Y-m-d', strtotime( $end_date ) ) === $end_date ) && ( '' !== $category_group ) ) {
						$table_name = $wpdb->prefix . 'smart_analytics_visitor_log';
						$query      = "SELECT IFNULL( NULLIF( RTRIM( LTRIM( $category_group ) ), '' ), 'Unknown' ) AS 'category_group', COUNT(DISTINCT name) AS 'visitor_count', COUNT(id) AS 'sum_page_load_count', SUM(time_in_sec) AS 'sum_duration_in_sec' FROM $table_name WHERE time  >= '$start_date' AND time <= '$end_date' $remove_bot_visits_str $remove_admin_visits_str GROUP BY category_group ORDER BY $sort $direction";
						$result     = $wpdb->get_results( $query, ARRAY_A );
						$count      = count( $result );

						for ( $i = 0; $i < $count; $i++ ) {
							$sno = $i + 1;
							$str = $str . '<tr>';
							$str = $str . '<td style="font-weight:bold; width:5%; overflow-x:hidden;">' . $sno . '</td>';
							$str = $str . '<td style="font-weight:bold; width:10%; min-width:20px; overflow-x:hidden;">' . $result[ $i ]['category_group'] . '</td>';
							$str = $str . '<td style="font-weight:bold; width:10%; min-width:20px; overflow-x:hidden;">' . $result[ $i ]['visitor_count'] . '</td>';
							$str = $str . '<td style="font-weight:bold; width:10%; min-width:20px; overflow-x:hidden;">' . $result[ $i ]['sum_page_load_count'] . '</td>';
							$str = $str . '<td style="font-weight:bold; width:10%; min-width:20px; overflow-x:hidden;">' . $result[ $i ]['sum_duration_in_sec'] . '</td>';
							$str = $str . '</tr>';
						}
					}
				}
			}

			echo $str;

			wp_die(); // Plugin is required to terminate immediately and return a proper response.
		}

		/**
		 * --------------------------------Shortcode Section-----------------------------------------.
		 */
		/**
		 * Registers shortcode assets.
		 *
		 * @return void
		 */
		public function register_shortcode() {
			add_shortcode( 'wpsa_log', array( $this, 'wpsa_log_shortcode_plugin' ) );
			add_shortcode( 'wpsa_count', array( $this, 'wpsa_count_shortcode_plugin' ) );

			add_action( 'wp_enqueue_scripts', array( $this, 'enqueue_shortcode_assets' ) );

			add_action( 'wp_ajax_wpsa_view_log_shortcode', array( $this, 'wp_ajax_wpsa_view_log_shortcode' ) );
			add_action( 'wp_ajax_nopriv_wpsa_view_log_shortcode', array( $this, 'wp_ajax_wpsa_view_log_shortcode' ) );
		}

		/**
		 * Registers shortcode assets.
		 *
		 * @return void
		 */
		public function enqueue_shortcode_assets() {
			global $post;
			if ( is_a( $post, 'WP_Post' ) && has_shortcode( $post->post_content, 'wpsa_log' ) ) {
				$min           = defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG ? '' : '.min';
				$my_js_url     = plugin_dir_url( __FILE__ ) . "assets/js/wpsa{$min}.js";
				$my_js_path    = plugin_dir_path( __FILE__ ) . "/assets/js/wpsa{$min}.js";
				$my_js_version = gmdate( 'ymd-Gis', filemtime( $my_js_path ) );

				$my_css_shortcode_url     = plugin_dir_url( __FILE__ ) . 'assets/css/wpsa-log.css';
				$my_css_shortcode_path    = plugin_dir_path( __FILE__ ) . '/assets/css/wpsa-log.css';
				$my_css_shortcode_version = gmdate( 'ymd-Gis', filemtime( $my_css_shortcode_path ) );

				$wpsc_js = wp_register_script( 'wpsa_js', $my_js_url, array( 'jquery', 'jquery-ui-core' ), $my_js_version, true );
				wp_localize_script(
					'wpsa_js',
					'ajax_var',
					array(
						'url'   => admin_url( 'admin-ajax.php' ),
						'nonce' => wp_create_nonce( 'wpsa_nonce' ),
					)
				);
				wp_enqueue_script( 'wpsa_js', $my_js_url, array( 'jquery', 'jquery-ui-core' ), $my_js_version, true );

				$wpsc_css_shortcode = wp_register_style( 'wpsa_css_shortcode', $my_css_shortcode_url, array(), $my_css_shortcode_version );
				wp_enqueue_style( 'wpsa_css_shortcode', $my_css_shortcode_url, array(), $my_css_shortcode_version );
			} elseif ( is_a( $post, 'WP_Post' ) && has_shortcode( $post->post_content, 'wpsa_count' ) ) {
				$my_css_shortcode_url     = plugin_dir_url( __FILE__ ) . 'assets/css/wpsa-count.css';
				$my_css_shortcode_path    = plugin_dir_path( __FILE__ ) . '/assets/css/wpsa-count.css';
				$my_css_shortcode_version = gmdate( 'ymd-Gis', filemtime( $my_css_shortcode_path ) );

				$wpsc_css_shortcode = wp_register_style( 'wpsa_css_shortcode', $my_css_shortcode_url, array(), $my_css_shortcode_version );
				wp_enqueue_style( 'wpsa_css_shortcode', $my_css_shortcode_url, array(), $my_css_shortcode_version );
			}
		}

		/**
		 * Shortcode Plugin wpsa_count.
		 * Sample [wpsa_count length=8 pad="0"  offset=100 style="color:#000; background-color:#fff; border-color:#000;"].
		 *
		 * @param array $atts Attributes of shortcode.
		 *
		 * @return string $str Output.
		 */
		public function wpsa_count_shortcode_plugin( $atts ) {
			$atts = shortcode_atts(
				array(
					'length' => 8,  // Length.
					'pad'    => '', // Pad character.
					'offset' => 0, // Add offset to the visitor count.
					'style'  => '', // Style.
				),
				$atts
			);

			$length = intval( $atts['length'] );
			if ( $length < 1 ) {
				$length = 8;
			}

			$pad = $atts['pad'];

			$offset = intval( $atts['offset'] );
			if ( $offset < 1 ) {
				$offset = 0;
			}

			$style = trim( $atts['style'] );
			$style = '' === $style ? '' : " style='" . $style . "'";

			$visitor_count = get_option( self::VISITOR_COUNT, 1 );

			$str = "<div id='wpsa_count'>";
			$str = $str . "<div class='wrap'>";
			$str = $str . "<div id='count'" . $style . '>';
			$str = $str . ( 0 === strlen( $pad ) ? $visitor_count + $offset : str_pad( $visitor_count + $offset, $length, $pad[0], STR_PAD_LEFT ) );
			$str = $str . '</div>';
			$str = $str . '</div>';
			$str = $str . '</div>';

			return $str;
		}

		/**
		 * Shortcode Plugin wpsa_log.
		 * Sample [wpsa_log style="color:#00a0c8; background-color:#fff;" heading_style="color:#00a0c8; background-color:#dbf3fa;"]
		 *
		 * @param array $atts Attributes of shortcode.
		 *
		 * @return string $str Output.
		 */
		public function wpsa_log_shortcode_plugin( $atts ) {
			$atts = shortcode_atts(
				array(
					'style'         => '', // Style.
					'heading_style' => '', // Heading Style.
				),
				$atts
			);

			$style = trim( $atts['style'] );
			$style = '' === $style ? '' : " style='" . $style . "'";

			$heading_style = trim( $atts['heading_style'] );
			$heading_style = '' === $heading_style ? '' : " style='" . $heading_style . "'";

			$str = "<div id='wpsa_log'>";
			$str = $str . "<script type='text/javascript'>let timer = null;</script>";
			$str = $str . "<div class='wrap'>";
			$str = $str . "<div id='log'" . $style . '>';
			$str = $str . '<h2 ' . $heading_style . '>' . __( 'Log Details.', 'wpsa' ) . '</h2>';
			$str = $str . $this->view_logs( 'shortcode' );
			$str = $str . '</div>';

			$str          = $str . "<div class='section_right'" . $style . '>';
			$str          = $str . '<h2 ' . $heading_style . '>' . __( 'Logs.', 'wpsa' ) . '</h2>';
			$str          = $str . '<p>' . __( 'Click to view relevant log details.', 'wpsa' ) . '</p>';
			$str          = $str . '<table>';
			$upload_dir   = wp_upload_dir();
			$wpsa_log_dir = trailingslashit( $upload_dir['basedir'] ) . '/wp-smart-analytics/log/';
			$files        = $this->get_files( $wpsa_log_dir );
			$cnt          = 0;
			foreach ( $files as $file ) {
				$cnt++;
				$str = $str . "<tr class='li' id='li_" . str_replace( '_', '', str_replace( '.', '', $file ) ) . "'>";
				$str = $str . "<td class='id'>{$cnt}</td><td class='filename' " . ( ( 1 === $cnt ) ? "style='font-weight:bold;'" : '' ) . '>' . str_replace( '_', ' ', str_replace( '.txt', '', $file ) ) . "</td><td class='action'><a onclick=\"ajax_view_log_shortcode('" . $file . "');\">VIEW</a></td>";
				$str = $str . '</tr>';
			}
			$str = $str . '</table>';
			$str = $str . '<div>&nbsp;</div>';
			$str = $str . "<div id='toTop'>To The Top!</div>";
			$str = $str . '<div>&nbsp;</div>';
			$str = $str . '</div>';
			$str = $str . '</div>';

			return $str;
		}

		/**
		 * Calls view_logs function. This is a ajax specific call which requires exit (ajax).
		 *
		 * @return void
		 */
		public function wp_ajax_wpsa_view_log_shortcode() {
			if ( isset( $_POST ) ) {
				$_POST = filter_input_array( INPUT_POST, FILTER_UNSAFE_RAW );
			}

			if ( wp_verify_nonce( isset( $_POST['nonce'] ) ? filter__sanitize_string_polyfill( filter_var( wp_unslash( $_POST['nonce'] ), FILTER_UNSAFE_RAW ) ) : '', 'wpsa_nonce' ) ) {
				echo $this->view_logs( 'read' );
			}

			wp_die(); // Plugin is required to terminate immediately and return a proper response.
		}

		/**
		 * --------------------------------Utility Section-----------------------------------------.
		 */
		/**
		 * Sanitizes the class name.
		 *
		 * @param string $class_name Class name.
		 *
		 * @return string $class_name Sanitized class name.
		 */
		public function sanitize_classname( $class_name ) {
			$pattern    = '/[^a-zA-Z0-9]/';
			$class_name = preg_replace( $pattern, '_', (string) $class_name );
			$class_name = ucfirst( $class_name );

			return $class_name;
		}

		/**
		 * Generates visitor name if does not exist.
		 *
		 * @param string $visitor_name Visitor name.
		 * @param string $replace_name Replacement name like Bingbot.
		 *
		 * @return string $visitor_name Visitor name.
		 */
		public function generate_visitorname( $visitor_name = '', $replace_name = '' ) {
			if ( isset( $visitor_name ) && ! empty( $visitor_name ) && ( false !== stripos( $visitor_name, '_' ) ) ) {
				$visitor_arr   = explode( '_', $visitor_name );
				$visitor_tag   = $visitor_arr[0];
				$visitor_count = $visitor_arr[1];
			} else {
				$visitor_tag   = 'Visitor';
				$visitor_count = get_option( self::VISITOR_COUNT, 1 );
				update_option( self::VISITOR_COUNT, $visitor_count + 1 );
			}

			if ( '' === $replace_name ) {
				$visitor_name = $visitor_tag . '_' . $visitor_count;
				if ( ( 'Visitor' === $visitor_tag ) && is_admin() ) {
					$visitor_name = 'Admin_' . $visitor_count; // Only visitor can be converted to admin.
				}
			} else {
				$visitor_name = ucwords( strtolower( $replace_name ) ) . '_' . $visitor_count; // If there is replacement proveided.
			}

			return $visitor_name;
		}

		/**
		 * Checks if string contains any value in substring array.
		 *
		 * @param string $str String.
		 * @param array  $sub_str_array Array of substrings.
		 *
		 * @return boolean True if string contains any  non whitespace value in substring array; otherwise false.
		 */
		public function check_substring( $str, $sub_str_array ) {
			foreach ( $sub_str_array as $sub_str ) {
				if ( '' !== trim( $sub_str ) ) {
					if ( stripos( $str, $sub_str ) !== false ) {
						return true;
					}
				}
			}
			return false;
		}

		/**
		 * Checks if ajax call.
		 *
		 * @return boolean True if AJAX call; Otherwise false.
		 */
		public function check_ajax() {
			if ( isset( $_SERVER['HTTP_X_REQUESTED_WITH'] ) && ! empty( $_SERVER['HTTP_X_REQUESTED_WITH'] ) ) {
				$request = filter_input( INPUT_SERVER, 'HTTP_X_REQUESTED_WITH', FILTER_UNSAFE_RAW );
				$request = strtolower( wp_unslash( $request ) );

				if ( 'xmlhttprequest' === $request ) {
					return true; // Handle request as AJAX.
				}
			}
			return false;
		}

		/**
		 * List down the files in the directory.
		 *
		 * @param string $path filepath.
		 *
		 * @return array $arr Sorted array of filenames in descending date value (from filename).
		 */
		public function get_files( $path ) {
			$arr      = array();
			$iterator = new FilesystemIterator( $path );
			foreach ( $iterator as $file_info ) {
				if ( $file_info->isFile() && ( '.htaccess' !== $file_info->getFileName() ) ) {
					$c_time = new DateTime();
					$c_time->setTimestamp( strtotime( str_replace( '_', ' ', str_replace( '.txt', '', $file_info->getFileName() ) ) ) );
					$arr[ $c_time->getTimestamp() ] = $file_info->getFileName();
				}
			}
			krsort( $arr );
			array_flip( $arr );

			return $arr;
		}

		/**
		 * Returns the IP Address of a Visitor.
		 *
		 * @param boolean $force_full_ip Forces return of full IP address.
		 *
		 * @return string $ip IP address.
		 */
		public function get_ip_address( $force_full_ip = false ) {
			$anonymize_ip = get_option( self::ANONYMIZE_IP, true );
			foreach ( array( 'HTTP_CLIENT_IP', 'HTTP_X_FORWARDED_FOR', 'HTTP_X_FORWARDED', 'HTTP_X_CLUSTER_CLIENT_IP', 'HTTP_FORWARDED_FOR', 'HTTP_FORWARDED', 'REMOTE_ADDR' ) as $key ) {
				if ( array_key_exists( $key, $_SERVER ) === true ) {
					$server = filter__sanitize_string_polyfill( filter_var( wp_unslash( $_SERVER[ $key ] ), FILTER_UNSAFE_RAW ) );
					foreach ( explode( ',', $server ) as $ip ) {
						if ( filter_var( $ip, FILTER_VALIDATE_IP ) !== false ) {
							$ip = trim( $ip );

							if ( true === $force_full_ip ) {
								return $ip;
							}

							if ( $anonymize_ip ) {
								if ( false !== strpos( $ip, '.' ) ) { // Detect IP type by dots instead of length.
									$pieces                    = explode( '.', $ip );
									$pieces_cnt                = count( $pieces );
									$pieces[ $pieces_cnt - 1 ] = preg_replace( '/\d/', 'X', $pieces[ $pieces_cnt - 1 ] );
									$ip                        = implode( '.', $pieces );
								} elseif ( false !== strpos( $ip, ':' ) ) { // Detect IP type by colons instead of length.
									// We need to expand the IPv6 to anonymize it properly.
									$hex = bin2hex( inet_pton( $ip ) );
									if ( '00000000000000000000ffff' === substr( $hex, 0, 24 ) ) { // IPv4-mapped IPv6 addresses.
										$ip = long2ip( hexdec( substr( $hex, -8 ) ) );
									}
									$ip = implode( ':', str_split( $hex, 4 ) );

									$pieces                    = explode( ':', $ip );
									$pieces_cnt                = count( $pieces );
									$pieces[ $pieces_cnt - 1 ] = preg_replace( '/\d/', 'X', $pieces[ $pieces_cnt - 1 ] );
									$pieces[ $pieces_cnt - 2 ] = preg_replace( '/\d/', 'X', $pieces[ $pieces_cnt - 2 ] );
									$pieces[ $pieces_cnt - 3 ] = preg_replace( '/\d/', 'X', $pieces[ $pieces_cnt - 3 ] );
									$pieces[ $pieces_cnt - 4 ] = preg_replace( '/\d/', 'X', $pieces[ $pieces_cnt - 4 ] );
									$pieces[ $pieces_cnt - 5 ] = preg_replace( '/\d/', 'X', $pieces[ $pieces_cnt - 5 ] );
									$ip                        = implode( ':', $pieces );
								}
							}

							return $ip;
						}
					}
				}
			}
		}

		/**
		 * Returns the URL accessed by Visitor.
		 *
		 * @return string $url URL.
		 */
		public function get_url() {
			global $wp;
			$url = add_query_arg( $wp->query_vars, home_url( $wp->request ) );

			return $url;
		}

		/**
		 * Returns the referrer URL of the Visitor if it is originating from a differenet host.
		 *
		 * @param string $referrer_url Referrer url.
		 *
		 * @return string $url URL.
		 */
		public function get_referrer( $referrer_url ) {
			if ( ! $referrer_url ) {
				return '';
			}

			$current_domain  = trim( str_replace( 'www.', '', strtolower( wp_parse_url( home_url(), PHP_URL_HOST ) ) ), '/' );
			$referrer_domain = trim( str_replace( 'www.', '', strtolower( wp_parse_url( $referrer_url, PHP_URL_HOST ) ) ), '/' );

			$current_port  = trim( wp_parse_url( home_url(), PHP_URL_PORT ) );
			$referrer_port = trim( wp_parse_url( $referrer_url, PHP_URL_PORT ) );

			if ( ( $current_domain === $referrer_domain ) && ( $current_port === $referrer_port ) ) {
				return '';
			}

			$referrer_url = trim( $referrer_url );
			return $referrer_url;
		}

		/**
		 * Returns the referrer source of the Visitor.
		 *
		 * @param string $referrer_url Referrer url.
		 *
		 * @return string $referrer_source Source.
		 */
		public function get_referrer_source( $referrer_url ) {
			if ( ! $referrer_url ) {
				return 'Unknown';
			}

			$current_domain  = trim( str_replace( 'www.', '', strtolower( wp_parse_url( home_url(), PHP_URL_HOST ) ) ), '/' );
			$referrer_domain = trim( str_replace( 'www.', '', strtolower( wp_parse_url( $referrer_url, PHP_URL_HOST ) ) ), '/' );

			$current_port  = trim( wp_parse_url( home_url(), PHP_URL_PORT ) );
			$referrer_port = trim( wp_parse_url( $referrer_url, PHP_URL_PORT ) );

			if ( ( $current_domain === $referrer_domain ) && ( $current_port === $referrer_port ) ) {
				return 'Internal';
			}

			if ( false !== strpos( $referrer_domain, '.' ) ) {
				$referrer_source = explode( '.', trim( $referrer_domain ) );
				array_pop( $referrer_source );
				return ucwords( array_pop( $referrer_source ) );
			}

			return 'External';
		}

		/**
		 * Returns the top preferred languages of visitor satisfying required priority.
		 *
		 * @param string $accept_languages Accept languages.
		 * @param string $priority Priority Minimum priority for a preferred language of visitor.
		 * @param string $limit Limit Number of preferred languages of visitor.
		 *
		 * @return string $url URL.
		 */
		public function get_user_base_language( $accept_languages, $priority, $limit ) {
			$languages_str = '';

			$cnt                = 0;
			$languages_code     = array();
			$languages_priority = array();
			if ( false !== preg_match_all( '/([a-z]{1,8})-[a-z]{1,8}\s*,[^;]*;\s*q\s*=\s*(1|0\.[0-9]+)/i', $accept_languages, $matches, PREG_SET_ORDER ) ) {
				foreach ( $matches as $key => $match ) {
					if ( ! in_array( trim( $match[1] ), $languages_code, true ) ) {
						if ( $match[2] >= $priority ) {
							$languages_code[ $cnt ]                        = trim( $match[1] );
							$languages_priority[ $languages_code[ $cnt ] ] = floatval( $match[2] );
							$cnt++;
						}
					}
				}
			}

			if ( $cnt < $limit ) {
				$limit = $cnt;
			}
			arsort( $languages_priority );

			$cnt = 0;
			foreach ( $languages_priority as $key => $language_priority ) {
				if ( $cnt < $limit - 1 ) {
					$languages_str = $this->get_language( $language_priority ) . ' ' . $languages_str;
				} else {
					$languages_str = $this->get_language( $language_priority ) . ' and ' . $languages_str;
					break;
				}
				$cnt++;
			}

			if ( 1 === $limit ) {
				$languages_str = $this->get_language( $languages_code[ $limit - 1 ] );
			}

			$languages_str = trim( $languages_str );
			return $languages_str;
		}

		/**
		 * Returns the language for a language code.
		 *
		 * @param string $language_code Language code.
		 *
		 * @return string $url URL.
		 */
		public function get_language( $language_code ) {
			$languages = array(
				'aa' => 'Afar',
				'ab' => 'Abkhazian',
				'ae' => 'Avestan',
				'af' => 'Afrikaans',
				'ak' => 'Akan',
				'am' => 'Amharic',
				'an' => 'Aragonese',
				'ar' => 'Arabic',
				'as' => 'Assamese',
				'av' => 'Avaric',
				'ay' => 'Aymara',
				'az' => 'Azerbaijani',
				'ba' => 'Bashkir',
				'be' => 'Belarusian',
				'bg' => 'Bulgarian',
				'bh' => 'Bihari',
				'bi' => 'Bislama',
				'bm' => 'Bambara',
				'bn' => 'Bengali',
				'bo' => 'Tibetan',
				'br' => 'Breton',
				'bs' => 'Bosnian',
				'ca' => 'Catalan',
				'ce' => 'Chechen',
				'ch' => 'Chamorro',
				'co' => 'Corsican',
				'cr' => 'Cree',
				'cs' => 'Czech',
				'cu' => 'Church Slavic',
				'cv' => 'Chuvash',
				'cy' => 'Welsh',
				'da' => 'Danish',
				'de' => 'German',
				'dv' => 'Divehi',
				'dz' => 'Dzongkha',
				'ee' => 'Ewe',
				'el' => 'Greek',
				'en' => 'English',
				'eo' => 'Esperanto',
				'es' => 'Spanish',
				'et' => 'Estonian',
				'eu' => 'Basque',
				'fa' => 'Persian',
				'ff' => 'Fulah',
				'fi' => 'Finnish',
				'fj' => 'Fijian',
				'fo' => 'Faroese',
				'fr' => 'French',
				'fy' => 'Western Frisian',
				'ga' => 'Irish',
				'gd' => 'Scottish Gaelic',
				'gl' => 'Galician',
				'gn' => 'Guarani',
				'gu' => 'Gujarati',
				'gv' => 'Manx',
				'ha' => 'Hausa',
				'he' => 'Hebrew',
				'hi' => 'Hindi',
				'ho' => 'Hiri Motu',
				'hr' => 'Croatian',
				'ht' => 'Haitian',
				'hu' => 'Hungarian',
				'hy' => 'Armenian',
				'hz' => 'Herero',
				'ia' => 'Interlingua (International Auxiliary Language Association)',
				'id' => 'Indonesian',
				'ie' => 'Interlingue',
				'ig' => 'Igbo',
				'ii' => 'Sichuan Yi',
				'ik' => 'Inupiaq',
				'io' => 'Ido',
				'is' => 'Icelandic',
				'it' => 'Italian',
				'iu' => 'Inuktitut',
				'ja' => 'Japanese',
				'jv' => 'Javanese',
				'ka' => 'Georgian',
				'kg' => 'Kongo',
				'ki' => 'Kikuyu',
				'kj' => 'Kwanyama',
				'kk' => 'Kazakh',
				'kl' => 'Kalaallisut',
				'km' => 'Khmer',
				'kn' => 'Kannada',
				'ko' => 'Korean',
				'kr' => 'Kanuri',
				'ks' => 'Kashmiri',
				'ku' => 'Kurdish',
				'kv' => 'Komi',
				'kw' => 'Cornish',
				'ky' => 'Kirghiz',
				'la' => 'Latin',
				'lb' => 'Luxembourgish',
				'lg' => 'Ganda',
				'li' => 'Limburgish',
				'ln' => 'Lingala',
				'lo' => 'Lao',
				'lt' => 'Lithuanian',
				'lu' => 'Luba-Katanga',
				'lv' => 'Latvian',
				'mg' => 'Malagasy',
				'mh' => 'Marshallese',
				'mi' => 'Maori',
				'mk' => 'Macedonian',
				'ml' => 'Malayalam',
				'mn' => 'Mongolian',
				'mr' => 'Marathi',
				'ms' => 'Malay',
				'mt' => 'Maltese',
				'my' => 'Burmese',
				'na' => 'Nauru',
				'nb' => 'Norwegian Bokmal',
				'nd' => 'North Ndebele',
				'ne' => 'Nepali',
				'ng' => 'Ndonga',
				'nl' => 'Dutch',
				'nn' => 'Norwegian Nynorsk',
				'no' => 'Norwegian',
				'nr' => 'South Ndebele',
				'nv' => 'Navajo',
				'ny' => 'Chichewa',
				'oc' => 'Occitan',
				'oj' => 'Ojibwa',
				'om' => 'Oromo',
				'or' => 'Oriya',
				'os' => 'Ossetian',
				'pa' => 'Panjabi',
				'pi' => 'Pali',
				'pl' => 'Polish',
				'ps' => 'Pashto',
				'pt' => 'Portuguese',
				'qu' => 'Quechua',
				'rm' => 'Raeto-Romance',
				'rn' => 'Kirundi',
				'ro' => 'Romanian',
				'ru' => 'Russian',
				'rw' => 'Kinyarwanda',
				'sa' => 'Sanskrit',
				'sc' => 'Sardinian',
				'sd' => 'Sindhi',
				'se' => 'Northern Sami',
				'sg' => 'Sango',
				'si' => 'Sinhala',
				'sk' => 'Slovak',
				'sl' => 'Slovenian',
				'sm' => 'Samoan',
				'sn' => 'Shona',
				'so' => 'Somali',
				'sq' => 'Albanian',
				'sr' => 'Serbian',
				'ss' => 'Swati',
				'st' => 'Southern Sotho',
				'su' => 'Sundanese',
				'sv' => 'Swedish',
				'sw' => 'Swahili',
				'ta' => 'Tamil',
				'te' => 'Telugu',
				'tg' => 'Tajik',
				'th' => 'Thai',
				'ti' => 'Tigrinya',
				'tk' => 'Turkmen',
				'tl' => 'Tagalog',
				'tn' => 'Tswana',
				'to' => 'Tonga',
				'tr' => 'Turkish',
				'ts' => 'Tsonga',
				'tt' => 'Tatar',
				'tw' => 'Twi',
				'ty' => 'Tahitian',
				'ug' => 'Uighur',
				'uk' => 'Ukrainian',
				'ur' => 'Urdu',
				'uz' => 'Uzbek',
				've' => 'Venda',
				'vi' => 'Vietnamese',
				'vo' => 'Volapuk',
				'wa' => 'Walloon',
				'wo' => 'Wolof',
				'xh' => 'Xhosa',
				'yi' => 'Yiddish',
				'yo' => 'Yoruba',
				'za' => 'Zhuang',
				'zh' => 'Chinese',
				'zu' => 'Zulu',
			);

			$language = '';

			if ( isset( $languages[ $language_code ] ) ) {
				$language = $languages[ $language_code ];
			}

			if ( ( version_compare( PHP_VERSION, '5.3.0', '>=' ) && class_exists( 'Locale' ) ) ) {
				$language_ret = locale_get_display_language( $language_code, 'en' );
				if ( $language_ret !== $language_code ) {
					$language = $language_ret;
				}
			}

			return ucwords( strtolower( $language ) );
		}

		/**
		 * Returns the type of WordPress page accessed by Visitor.
		 * Based on https://wordpress.stackexchange.com/questions/83887/return-current-page-type by Eugene Manuilov.
		 *
		 * @return string $type Type.
		 */
		public function get_page_type() {
			global $wp_query;

			$type = '';

			if ( $wp_query->is_page ) {
				$type = is_front_page() ? 'front page' : 'page';
			} elseif ( $wp_query->is_home ) {
				$type = 'home page';
			} elseif ( $wp_query->is_single ) {
				$type = ( $wp_query->is_attachment ) ? 'single attachment post' : 'single post';
			} elseif ( $wp_query->is_category ) {
				$type = 'category page';
			} elseif ( $wp_query->is_tag ) {
				$type = 'tag';
			} elseif ( $wp_query->is_tax ) {
				$type = 'taxomony';
			} elseif ( $wp_query->is_archive ) {
				if ( $wp_query->is_day ) {
					$type = 'day archive';
				} elseif ( $wp_query->is_month ) {
					$type = 'month archive';
				} elseif ( $wp_query->is_year ) {
					$type = 'year archive';
				} elseif ( $wp_query->is_author ) {
					$type = 'author profile';
				} else {
					$type = 'archive';
				}
			} elseif ( $wp_query->is_search ) {
				$type = 'search page';
			} elseif ( $wp_query->is_404 ) {
				$type = '404 page';
			}

			$type = trim( $type );
			return $type;
		}

		/**
		 * Returns the title of page accessed by Visitor.
		 *
		 * @return string $title Title.
		 */
		public function get_title() {
			$title = wp_title( '', false, 'right' );

			if ( ! $title ) {
				$title = '';
			}

			$title = trim( $title );
			return $title;
		}

		/**
		 * Returns the current time.
		 *
		 * @return string $current_time Currrent time.
		 */
		public function get_current_time() {
			$timestamp    = time();
			$current_time = gmdate( 'Y-m-d\TH:i:s\Z', $timestamp );
			return $current_time;
		}

		/**
		 * Get the color code for a string.
		 * Generates background color code.
		 *
		 * @param string $str Input string.
		 *
		 * @return string $color Color code.
		 */
		public function get_lightcolor( $str ) {
			$slack = self::SLACK; // Slack ensures the color is not too light.

			$str = str_pad( $str, 3 );

			$str_length = strlen( $str );
			$parts      = floor( $str_length / 3 ); // Break string into 3 parts.
			$str_chunks = chunk_split( $str, $parts );
			$str_arr    = array_filter( explode( PHP_EOL, $str_chunks ) );

			$colornum_r = hexdec( substr( strrev( md5( $str_arr[0] ) ), 0, 2 ) );
			$colornum_r = 128 + $colornum_r / 2 - $slack;
			$colornum_r = dechex( $colornum_r );

			$colornum_g = hexdec( substr( strrev( md5( $str_arr[1] ) ), 0, 2 ) );
			$colornum_g = 128 + $colornum_g / 2 - $slack;
			$colornum_g = dechex( $colornum_g );

			$colornum_b = hexdec( substr( strrev( md5( $str_arr[2] ) ), 0, 2 ) );
			$colornum_b = 128 + $colornum_b / 2 - $slack;
			$colornum_b = dechex( $colornum_b );

			$color = '#' . $colornum_r . $colornum_g . $colornum_b;
			return $color;
		}

		/**
		 * Get the color code for a string.
		 * Generates foreground color code.
		 *
		 * @param string $str Input string.
		 *
		 * @return string $color Color code.
		 */
		public function get_darkcolor( $str ) {
			$slack = self::SLACK; // Slack ensures the color is not too dark.

			$str = str_pad( $str, 3 );

			$str_length = strlen( $str );
			$parts      = floor( $str_length / 3 ); // Break string into 3 parts.
			$str_chunks = chunk_split( $str, $parts );
			$str_arr    = array_filter( explode( PHP_EOL, $str_chunks ) );

			$colornum_r = hexdec( substr( strrev( md5( $str_arr[0] ) ), 0, 2 ) );
			$colornum_r = 128 - $colornum_r / 2 + $slack;
			$colornum_r = dechex( $colornum_r );

			$colornum_g = hexdec( substr( strrev( md5( $str_arr[1] ) ), 0, 2 ) );
			$colornum_g = 128 - $colornum_g / 2 + $slack;
			$colornum_g = dechex( $colornum_g );

			$colornum_b = hexdec( substr( strrev( md5( $str_arr[2] ) ), 0, 2 ) );
			$colornum_b = 128 - $colornum_b / 2 + $slack;
			$colornum_b = dechex( $colornum_b );

			$color = '#' . $colornum_r . $colornum_g . $colornum_b;
			return $color;
		}

		/**
		 * Encryption / decryption function using SHA 256.
		 *
		 * @param string $str Encryption or decrypttion string depending on action.
		 * @param string $action Encryption or decryption action. Defaults to encryption.
		 *
		 * @return string $output Encryption or decrypttion string depending on action.
		 */
		public function encrypt_decrypt( $str, $action = 'encrypt' ) {
			$encrypt_method = 'AES-256-CBC';
			$secret_key     = get_option( self::SHA_SECRET_KEY, substr( str_shuffle( '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ' ), 1, 32 ) ); // User defined private key.
			$secret_iv      = get_option( self::SHA_SECRET_IV, substr( str_shuffle( '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ' ), 1, 32 ) ); // User defined secret key.
			$key            = hash( 'sha256', $secret_key );
			$iv             = substr( hash( 'sha256', $secret_iv ), 0, 16 ); // sha256 is hash_hmac_algorithm.
			if ( 'encrypt' === $action ) {
				$output = openssl_encrypt( $str, $encrypt_method, $key, 0, $iv );
				$output = base64_encode( $output );
			} elseif ( 'decrypt' === $action ) {
				$output = openssl_decrypt( base64_decode( $str ), $encrypt_method, $key, 0, $iv );
			}

			return $output;
		}

		/**
		 * Makes duration verbose
		 *
		 * @param string $duration Duration.
		 *
		 * @return string $verbose Verbose.
		 */
		public function duration_verbose( $duration ) {
			// Return type  is float.
			$days                      = intval( floor( $duration / ( 24 * 60 * 60 ) ) );
			$hours_and_minutes_seconds = intval( fmod( $duration, 24 * 60 * 60 ) );
			$hours                     = intval( floor( $hours_and_minutes_seconds / ( 60 * 60 ) ) );
			$minutes_and_seconds       = intval( fmod( $hours_and_minutes_seconds, 60 * 60 ) );
			$minutes                   = intval( floor( $minutes_and_seconds / 60 ) );
			$seconds                   = intval( fmod( $minutes_and_seconds, 60 ) );

			$days_str    = $days > 0 ? ( 1 === $days ? '1 day' : $days . ' days' ) : '';
			$hours_str   = $hours > 0 ? ( 1 === $hours ? '1 hour' : $hours . ' hours' ) : '';
			$minutes_str = $minutes > 0 ? ( 1 === $minutes ? '1 minute' : $minutes . ' minutes' ) : '';
			$seconds_str = $seconds > 0 ? ( 1 === $seconds ? '1 second' : $seconds . ' seconds' ) : '';

			$verbose = '';
			if ( $days > 0 ) {
				$verbose = $verbose . $days_str;
			}
			if ( $hours > 0 ) {
				$verbose = $verbose . ' and ' . $hours_str;
			}
			if ( $minutes > 0 ) {
				$verbose = $verbose . ' and ' . $minutes_str;
			}
			if ( $seconds > 0 ) {
				$verbose = $verbose . ' and ' . $seconds_str;
			}

			$verbose = ltrim( $verbose, ' and ' );

			return $verbose;
		}

		/**
		 * Check if frontend.
		 *
		 * @return boolean Returns true if front end page; else false.
		 */
		public function is_frontend() {
			$path_array = str_replace( array( '\\', '/' ), DIRECTORY_SEPARATOR, ABSPATH );
			$php_self   = filter_input( INPUT_SERVER, 'PHP_SELF', FILTER_SANITIZE_URL );
			return ( ! ( is_admin() || wp_doing_ajax() || in_array( $path_array . 'wp-login.php', get_included_files(), true ) || in_array( $path_array . 'wp-register.php', get_included_files(), true ) ) || 'wp-login.php' === $GLOBALS['pagenow'] || '/wp-login.php' === $php_self );
		}

		/**
		 * Get location details from IP.
		 * ip "12.12.12.12".
		 * country_code "US".
		 * country_name "United States".
		 * region_name  "Alaska".
		 * city "Anchorage".
		 * zip  "99501".
		 *
		 * @param string $ip IP.
		 *
		 * @return string $api_result JSON with location  details.
		 */
		public function get_ip_location_details( $ip ) {
			$ip = str_replace( 'X', '0', $ip );

			try {
				$attempt = self::MAX_READ_URL; // Defaults to 3. Should be >= 1.
				do {
					// Fetches the data.
					$request = wp_remote_get(
						self::ACCESS_URL . $ip . '?access_key=' . get_option( self::WPSA_API_KEY, '' ),
						array(
							'timeout' => self::MAX_READ_TIMEOUT,
						)
					);
					--$attempt;
				} while ( is_wp_error( $request ) && ( $attempt > 0 ) );

				// Decode JSON response.
				$response   = wp_remote_retrieve_body( $request );
				$api_result = json_decode( $response, true );
			} catch ( Exception $e ) {
				$api_result = '';
			}

			return $api_result;
		}

		/**
		 * Gets the Bot information.
		 * Based on work found in https://www.php.net/manual/en/function.get-browser.php attributed to Anonymous.
		 *
		 * @param string $user_agent User agent.
		 *
		 * @return string $bot bot information.
		 */
		public function bot_info( $user_agent ) {
			// Search Engines.
			if ( 1 === preg_match( '/google/i', $user_agent, $matches ) ) {
				return 'Google';
			} elseif ( 1 === preg_match( '/bing/i', $user_agent, $matches ) ) {
				return 'Bing';
			} elseif ( 1 === preg_match( '/slurp/i', $user_agent, $matches ) ) {
				return 'Yahoo Slurp';
			} elseif ( 1 === preg_match( '/duckduckgo/i', $user_agent, $matches ) ) {
				return 'DuckDuckBot';
			} elseif ( 1 === preg_match( '/baidu/i', $user_agent, $matches ) ) {
				return 'Baidu';
			} elseif ( 1 === preg_match( '/yandex/i', $user_agent, $matches ) ) {
				return 'Yandex';
			} elseif ( 1 === preg_match( '/sogou/i', $user_agent, $matches ) ) {
				return 'Sogou';
			} elseif ( 1 === preg_match( '/exabot/i', $user_agent, $matches ) ) {
				return 'Exabot';
			} elseif ( 1 === preg_match( '/msn/i', $user_agent, $matches ) ) {
				return 'MSN';
			}

			// Common Tools and Bots.
			if ( 1 === preg_match( '/mj12bot/i', $user_agent, $matches ) ) {
				return 'Majestic';
			} elseif ( 1 === preg_match( '/ahrefs/i', $user_agent, $matches ) ) {
				return 'Ahrefs';
			} elseif ( 1 === preg_match( '/semrush/i', $user_agent, $matches ) ) {
				return 'SEMRush';
			} elseif ( 1 === preg_match( '/rogerbot|dotbot/i', $user_agent, $matches ) ) {
				return 'Moz or OpenSiteExplorer';
			} elseif ( 1 === preg_match( '/frog|screaming/i', $user_agent, $matches ) ) {
				return 'Screaming Frog';
			}

			// Miscellaneous.
			if ( 1 === preg_match( '/facebook/i', $user_agent, $matches ) ) {
				return 'Facebook';
			} elseif ( 1 === preg_match( '/pininterest/i', $user_agent, $matches ) ) {
				return 'Pinterest';
			}

			// Check for strings commonly used in bot user agents.
			if ( 1 === preg_match( '/crawler|api|spider|http|bot|archive|info|data|nuhk/i', $user_agent, $matches ) ) {
				return 'Other';
			}

			return false;
		}

		/**
		 * Gets the Operating system information.
		 * Based on work found in stackoverflow attributed to Daniel Kassner.
		 *
		 * @param string $user_agent User agent.
		 *
		 * @return string $os Operating system information.
		 */
		public function os_info( $user_agent ) {
			// The order of this array is important.
			$oses = array(
				'Win311'        => 'Win16',
				'Win95'         => 'Windows 95',
				'Win95'         => 'Win95',
				'Win95'         => 'Windows_95',
				'WinME'         => 'Windows 98',
				'WinME'         => 'Win 9x 4.90',
				'WinME'         => 'Windows ME',
				'Win98'         => 'Windows 98',
				'Win98'         => 'Win98',
				'Win2000'       => 'Windows NT 5.0',
				'Win2000'       => 'Windows 2000',
				'WinXP'         => 'Windows NT 5.1',
				'WinXP'         => 'Windows XP',
				'WinServer2003' => 'Windows NT 5.2',
				'WinVista'      => 'Windows NT 6.0',
				'Windows 7'     => 'Windows NT 6.1',
				'Windows 8'     => 'Windows NT 6.2',
				'Windows 8.1'   => 'Windows NT 6.3',
				'Windows 10'    => 'Windows NT 10.0',
				'WinNT'         => 'Windows NT 4.0',
				'WinNT'         => 'WinNT4.0',
				'WinNT'         => 'WinNT',
				'WinNT'         => 'Windows NT',
				'OpenBSD'       => 'OpenBSD',
				'SunOS'         => 'SunOS',
				'Ubuntu'        => 'Ubuntu',
				'Android'       => 'Android',
				'Linux'         => 'Linux',
				'Linux'         => 'X11',
				'iPhone'        => 'iPhone',
				'iPad'          => 'iPad',
				'MacOS'         => 'Mac_PowerPC',
				'MacOS'         => 'Macintosh',
				'QNX'           => 'QNX',
				'BeOS'          => 'BeOS',
				'OS2'           => 'OS/2',
			);

			foreach ( $oses as $os => $pattern ) {
				if ( false !== stripos( $user_agent, $pattern ) ) {
					return $os;
				}
			}

			return 'Unknown';
		}

		/**
		 * Gets the Browser information.
		 * Based on work found in https://www.php.net/manual/en/function.get-browser.php attributed to Shashank.
		 *
		 * @param string $user_agent User agent.
		 * @param string $type Provides browser info for $type ='browser' and version information for type = 'version'. Gives both informations if not set.
		 *
		 * @return string $full_browser_info Browser Information.
		 */
		public function browser_info( $user_agent, $type = '' ) {
			$bname = 'Unknown';
			$ub    = '';

			if ( preg_match( '/Opera|OPR\//i', $user_agent ) ) {
				$bname = 'Opera';
				$ub    = 'Opera';
			} elseif ( preg_match( '/Edge/i', $user_agent ) ) {
				$bname = 'Edge';
				$ub    = 'Edge';
			} elseif ( preg_match( '/Chrome/i', $user_agent ) ) {
				$bname = 'Google Chrome';
				$ub    = 'Chrome';
			} elseif ( preg_match( '/Safari/i', $user_agent ) ) {
				$bname = 'Apple Safari';
				$ub    = 'Safari';
			} elseif ( preg_match( '/Firefox/i', $user_agent ) ) {
				$bname = 'Mozilla Firefox';
				$ub    = 'Firefox';
			} elseif ( preg_match( '/MSIE|Trident\/7/i', $user_agent ) ) {
				$bname = 'Internet Explorer';
				$ub    = 'MSIE';
			} elseif ( preg_match( '/Netscape/i', $user_agent ) ) {
				$bname = 'Netscape';
				$ub    = 'Netscape';
			}

			if ( 'browser' === $type ) {
				return $bname;
			}

			// Finally get the correct version number.
			$known   = array( 'Version', $ub, 'other' );
			$pattern = '#(?<browser>' . join( '|', $known ) . ')[/ ]+(?<version>[0-9.|a-zA-Z.]*)#';
			if ( false !== preg_match_all( $pattern, $user_agent, $matches ) ) {
				// See how many we have.
				$i = count( $matches['browser'] );
			}
			if ( 1 !== $i ) {
				// We will have two since we are not using 'other' argument yet.
				// See if version is before or after the name.
				if ( strripos( $user_agent, 'Version' ) < strripos( $user_agent, $ub ) ) {
					$version = $matches['version'][0];
				} else {
					$version = $matches['version'][1];
				}
			} else {
				$version = $matches['version'][0];
			}

			// Check if we have a number.
			if ( null === $version || '' === $version ) {
				$version = 'Unknown';
			}

			if ( 'version' === $type ) {
				return $version;
			}

			$full_browser_info = $bname . ' browser of version ' . $version;

			return $full_browser_info;
		}

		/**
		 * Gets required visitor location from ip location details.
		 *
		 * @param array $ip_location IP location details.
		 *
		 * @return string $visitor_location Visitor location.
		 */
		public function get_visitor_location_from_ip_location_details( $ip_location ) {
			$zip          = '';
			$city         = '';
			$region_name  = '';
			$country_name = '';

			if ( isset( $ip_location['city'] ) ) {
				$city = $ip_location['city'];
			}
			if ( isset( $ip_location['region_name'] ) ) {
				$region_name = $ip_location['region_name'];
			}
			if ( isset( $ip_location['zip'] ) ) {
				$zip = $ip_location['zip'];
			}
			if ( isset( $ip_location['country_name'] ) ) {
				$country_name = $ip_location['country_name'];
			}

			if ( strlen( $zip ) > 0 ) {
				$zip = $zip . ', ';
			}
			if ( strlen( $city ) > 0 ) {
				$city = $city . ', ';
			}
			if ( strlen( $region_name ) > 0 ) {
				$region_name = $region_name . ', ';
			}
			if ( strlen( $country_name ) > 0 ) {
				$country_name = $country_name;
			}

			$visitor_location = trim( $zip . $city . $region_name . $country_name );

			return $visitor_location;
		}
	}

	/**
	 * Polyfill for FILTER_SANITIZE_STRING.
	 *
	 * @param string $string Input.
	 *
	 * @return string Output.
	 */
	function filter__sanitize_string_polyfill( string $string ): string {
		$str = preg_replace( '/\x00|<[^>]*>?/', '', $string );
		return str_replace( array( "'", '"' ), array( '&#39;', '&#34;' ), $str );
	}

	// Initialize.
	WPSA::get_instance();
endif;
